#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.session
   :platform: Unix
   :synopsis: Implementation of the Session class.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os 
import glob
import saga
import radical.utils as ru

from radical.pilot.object          import Object
from radical.pilot.unit_manager    import UnitManager
from radical.pilot.pilot_manager   import PilotManager
from radical.pilot.utils.logger    import logger
from radical.pilot.resource_config import ResourceConfig
from radical.pilot.exceptions      import PilotException

from radical.pilot.db              import Session as dbSession
from radical.pilot.db              import DBException

from bson.objectid                 import ObjectId



# ------------------------------------------------------------------------------
#
class _ProcessRegistry(object):
    """A _ProcessRegistry contains a dictionary of all worker processes 
    that are currently active.
    """
    def __init__(self):
        self._dict = dict()

    def register(self, key, process):
        """Add a new process to the registry.
        """
        if key not in self._dict:
            self._dict[key] = process

    def retrieve(self, key):
        """Retrieve a process from the registry.
        """
        if key not in self._dict:
            return None
        else:
            return self._dict[key]

    def keys(self):
        """List all keys of all process in the registry.
        """
        return self._dict.keys()

    def remove(self, key):
        """Remove a process from the registry.
        """
        if key in self._dict:
            del self._dict[key]

# ------------------------------------------------------------------------------
#
class Session (saga.Session, Object):
    """A Session encapsulates a RADICAL-Pilot instance and is the *root* object
    for all other RADICAL-Pilot objects. 

    A Session holds :class:`radical.pilot.PilotManager` and :class:`radical.pilot.UnitManager`
    instances which in turn hold  :class:`radical.pilot.Pilot` and
    :class:`radical.pilot.ComputeUnit` instances.

    Each Session has a unique identifier :data:`radical.pilot.Session.uid` that can be
    used to re-connect to a RADICAL-Pilot instance in the database.

    **Example**::

        s1 = radical.pilot.Session(database_url=DBURL)
        s2 = radical.pilot.Session(database_url=DBURL, session_uid=s1.uid)

        # s1 and s2 are pointing to the same session
        assert s1.uid == s2.uid
    """

    #---------------------------------------------------------------------------
    #
    def __init__ (self, database_url=None, database_name="radicalpilot", session_uid=None):
        """Creates a new or reconnects to an exising session.

        If called without a session_uid, a new Session instance is created and 
        stored in the database. If session_uid is set, an existing session is 
        retrieved from the database. 

        **Arguments:**
            * **database_url** (`string`): The MongoDB URL.  If none is given,
              RP uses the environment variable RADICAL_PILOT_DBURL.  If that is
              not set, an error will be raises.

            * **database_name** (`string`): An alternative database name 
              (default: 'radicalpilot').

            * **session_uid** (`string`): If session_uid is set, we try 
              re-connect to an existing session instead of creating a new one.

        **Returns:**
            * A new Session instance.

        **Raises:**
            * :class:`radical.pilot.DatabaseError`

        """

        # init the base class inits
        saga.Session.__init__ (self)
        Object.__init__ (self)

        # before doing anything else, set up the debug helper for the lifetime
        # of the session.
        self._debug_helper = ru.DebugHelper ()

        # Dictionaries holding all manager objects created during the session.
        self._pilot_manager_objects = list()
        self._unit_manager_objects = list()

        # Create a new process registry. All objects belonging to this 
        # session will register their worker processes (if they have any)
        # in this registry. This makes it easier to shut down things in 
        # a more coordinate fashion. 
        self._process_registry = _ProcessRegistry()

        # The resource configuration dictionary associated with the session.
        self._resource_configs = {}

        self._database_url  = database_url
        self._database_name = database_name 

        if  not self._database_url :
            self._database_url = os.getenv ("RADICAL_PILOT_DBURL", None)

        if  not self._database_url :
            raise PilotException ("no database URL (set RADICAL_PILOT_DBURL)")  

        logger.info("using database url  %s" % self._database_url)

        # if the database url contains a path element, we interpret that as
        # database name (without the leading slash)
        tmp_url = ru.Url (self._database_url)
        if  tmp_url.path            and \
            tmp_url.path[0]  == '/' and \
            len(tmp_url.path) >  1  :
            self._database_name = tmp_url.path[1:]
            logger.info("using database path %s" % self._database_name)
        else :
            logger.info("using database name %s" % self._database_name)

        # Loading all "default" resource configurations
        module_path   = os.path.dirname(os.path.abspath(__file__))
        default_cfgs  = "%s/configs/*.json" % module_path
        config_files  = glob.glob(default_cfgs)

        for config_file in config_files:
            rcs = ResourceConfig.from_file(config_file)

            if  rcs :
                for rc in rcs:
                    logger.info("Loaded resource configurations for %s" % rc)
                    self._resource_configs[rc] = rcs[rc].as_dict() 

        user_cfgs     = "%s/.radical/pilot/configs/*.json" % os.environ.get ('HOME')
        config_files  = glob.glob(user_cfgs)

        for config_file in config_files:
            rcs = ResourceConfig.from_file(config_file)

            if  rcs :
                for rc in rcs:
                    logger.info("Loaded resource configurations for %s" % rc)

                    if  rc in self._resource_configs :
                        # config exists -- merge user config into it
                        ru.dict_merge (self._resource_configs[rc],
                                       rcs[rc].as_dict(),
                                       policy='overwrite')
                    else :
                        # new config -- add as is
                        self._resource_configs[rc] = rcs[rc].as_dict() 

        default_aliases = "%s/configs/aliases.json" % module_path
        self._resource_aliases = ru.read_json_str (default_aliases)['aliases']

        ##########################
        ## CREATE A NEW SESSION ##
        ##########################
        if session_uid is None:
            try:
                self._uid = str(ObjectId())
                self._last_reconnect = None

                self._dbs, self._created, self._connection_info = \
                        dbSession.new(sid=self._uid,
                                      db_url=self._database_url,
                                      db_name=database_name)

                logger.info("New Session created%s." % str(self))

            except Exception, ex:
                raise PilotException("Couldn't create new session (database URL '%s' incorrect?): %s" \
                                % (self._database_url, ex))  

        ######################################
        ## RECONNECT TO AN EXISTING SESSION ##
        ######################################
        else:
            try:
                self._uid = session_uid

                # otherwise, we reconnect to an existing session
                self._dbs, session_info, self._connection_info = \
                        dbSession.reconnect(sid=self._uid, 
                                            db_url=self._database_url,
                                            db_name=database_name)

                self._created          = session_info["created"]
                self._last_reconnect   = session_info["last_reconnect"]

                logger.info("Reconnected to existing Session %s." % str(self))

            except Exception, ex:
                raise PilotException("Couldn't re-connect to session: %s" % ex)  

    #---------------------------------------------------------------------------
    #
    def close(self, cleanup=True, terminate=True, delete=None):
        """Closes the session.

        All subsequent attempts access objects attached to the session will 
        result in an error. If cleanup is set to True (default) the session
        data is removed from the database.

        **Arguments:**
            * **cleanup** (`bool`): Remove session from MongoDB (implies * terminate)
            * **terminate** (`bool`): Shut down all pilots associated with the session. 

        **Raises:**
            * :class:`radical.pilot.IncorrectState` if the session is closed
              or doesn't exist. 
        """

        uid = self._uid

        if not self._uid:
            logger.warning("Session object already closed.")
            return

        # we keep 'delete' for backward compatibility.  If it was set, and the
        # other flags (cleanup, terminate) are as defaulted (True), then delete
        # will supercede them.  Delete is considered deprecated though, and
        # we'll thus issue a warning.
        if  delete != None:

            if  cleanup == True and terminate == True :
                cleanup   =  delete
                terminate = delete
                logger.warning("'delete' flag on session is deprecated. " \
                               "Please use 'cleanup' and 'terminate' instead!")

        if  cleanup :
            # cleanup implies terminate
            terminate = True

        if terminate :
            # cancel all pilots, make sure they are gone, and close the pilot
            # managers.
            for pmgr in self._pilot_manager_objects:
                pmgr.cancel_pilots ()

            for pmgr in self._pilot_manager_objects:
                pmgr.wait_pilots ()

            for pmgr in self._pilot_manager_objects:
                pmgr.close ()

        for umgr in self._unit_manager_objects:
            umgr.close()

        if  cleanup :
            self._destroy_db_entry()

        logger.info("Closed Session %s." % str(uid))


    #---------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns a Python dictionary representation of the object.
        """
        object_dict = {
            "uid": self._uid,
            "created": self._created,
            "last_reconnect": self._last_reconnect,
            "database_name": self._connection_info.dbname,
            "database_auth": self._connection_info.dbauth,
            "database_url" : self._connection_info.dburl
        }
        return object_dict

    #---------------------------------------------------------------------------
    #
    def __str__(self):
        """Returns a string representation of the object.
        """
        return str(self.as_dict())

    #---------------------------------------------------------------------------
    #
    @property
    def created(self):
        """Returns the UTC date and time the session was created.
        """
        self._assert_obj_is_valid()
        return self._created

    #---------------------------------------------------------------------------
    #
    @property
    def last_reconnect(self):
        """Returns the most recent UTC date and time the session was
        reconnected to.
        """
        self._assert_obj_is_valid()
        return self._last_reconnect


    #---------------------------------------------------------------------------
    #
    def _destroy_db_entry(self):
        """Terminates the session and removes it from the database.

        All subsequent attempts access objects attached to the session and 
        attempts to re-connect to the session via its uid will result in
        an error.

        **Raises:**
            * :class:`radical.pilot.IncorrectState` if the session is closed
              or doesn't exist. 
        """
        self._assert_obj_is_valid()

        self._dbs.delete()
        logger.info("Deleted session %s from database." % self._uid)
        self._uid = None

    #---------------------------------------------------------------------------
    #
    def list_pilot_managers(self):
        """Lists the unique identifiers of all :class:`radical.pilot.PilotManager` 
        instances associated with this session.

        **Example**::

            s = radical.pilot.Session(database_url=DBURL)
            for pm_uid in s.list_pilot_managers():
                pm = radical.pilot.PilotManager(session=s, pilot_manager_uid=pm_uid) 

        **Returns:**
            * A list of :class:`radical.pilot.PilotManager` uids (`list` oif strings`).

        **Raises:**
            * :class:`radical.pilot.IncorrectState` if the session is closed
              or doesn't exist. 
        """
        self._assert_obj_is_valid()
        return self._dbs.list_pilot_manager_uids()


    # --------------------------------------------------------------------------
    #
    def get_pilot_managers(self, pilot_manager_ids=None) :
        """ Re-connects to and returns one or more existing PilotManager(s).

        **Arguments:**

            * **session** [:class:`radical.pilot.Session`]: 
              The session instance to use.

            * **pilot_manager_uid** [`string`]: 
              The unique identifier of the PilotManager we want 
              to re-connect to.

        **Returns:**

            * One or more new [:class:`radical.pilot.PilotManager`] objects.

        **Raises:**

            * :class:`radical.pilot.pilotException` if a PilotManager with 
              `pilot_manager_uid` doesn't exist in the database.
        """
        self._assert_obj_is_valid()

        return_scalar = False

        if pilot_manager_ids is None:
            pilot_manager_ids = self.list_pilot_managers()

        elif not isinstance(pilot_manager_ids, list):
            pilot_manager_ids = [pilot_manager_ids]
            return_scalar = True

        pilot_manager_objects = []

        for pilot_manager_id in pilot_manager_ids:
            pilot_manager = PilotManager._reconnect(session=self, pilot_manager_id=pilot_manager_id)
            pilot_manager_objects.append(pilot_manager)

            self._pilot_manager_objects.append(pilot_manager)

        if return_scalar is True:
            pilot_manager_objects = pilot_manager_objects[0]

        return pilot_manager_objects

    #---------------------------------------------------------------------------
    #
    def list_unit_managers(self):
        """Lists the unique identifiers of all :class:`radical.pilot.UnitManager` 
        instances associated with this session.

        **Example**::

            s = radical.pilot.Session(database_url=DBURL)
            for pm_uid in s.list_unit_managers():
                pm = radical.pilot.PilotManager(session=s, pilot_manager_uid=pm_uid) 

        **Returns:**
            * A list of :class:`radical.pilot.UnitManager` uids (`list` of `strings`).

        **Raises:**
            * :class:`radical.pilot.IncorrectState` if the session is closed
              or doesn't exist. 
        """
        self._assert_obj_is_valid()
        return self._dbs.list_unit_manager_uids()

    # --------------------------------------------------------------------------
    #
    def get_unit_managers(self, unit_manager_ids=None) :
        """ Re-connects to and returns one or more existing UnitManager(s).

        **Arguments:**

            * **session** [:class:`radical.pilot.Session`]: 
              The session instance to use.

            * **pilot_manager_uid** [`string`]: 
              The unique identifier of the PilotManager we want 
              to re-connect to.

        **Returns:**

            * One or more new [:class:`radical.pilot.PilotManager`] objects.

        **Raises:**

            * :class:`radical.pilot.pilotException` if a PilotManager with 
              `pilot_manager_uid` doesn't exist in the database.
        """
        self._assert_obj_is_valid()

        return_scalar = False
        if unit_manager_ids is None:
            unit_manager_ids = self.list_unit_managers()

        elif not isinstance(unit_manager_ids, list):
            unit_manager_ids = [unit_manager_ids]
            return_scalar = True

        unit_manager_objects = []

        for unit_manager_id in unit_manager_ids:
            unit_manager = UnitManager._reconnect(session=self, unit_manager_id=unit_manager_id)
            unit_manager_objects.append(unit_manager)

            self._unit_manager_objects.append(unit_manager)

        if return_scalar is True:
            unit_manager_objects = unit_manager_objects[0]

        return unit_manager_objects

    # -------------------------------------------------------------------------
    #
    def add_resource_config(self, resource_config):
        """Adds a new :class:`radical.pilot.ResourceConfig` to the PilotManager's 
           dictionary of known resources, or accept a string which points to
           a configuration file.

           For example::

                  rc = radical.pilot.ResourceConfig
                  rc.name                 = "mycluster"
                  rc.job_manager_endpoint = "ssh+pbs://mycluster
                  rc.filesystem_endpoint  = "sftp://mycluster
                  rc.default_queue        = "private"
                  rc.bootstrapper         = "default_bootstrapper.sh"

                  pm = radical.pilot.PilotManager(session=s)
                  pm.add_resource_config(rc)

                  pd = radical.pilot.ComputePilotDescription()
                  pd.resource = "mycluster"
                  pd.cores    = 16
                  pd.runtime  = 5 # minutes

                  pilot = pm.submit_pilots(pd)
        """
        if  isinstance (resource_config, basestring) :

            rcs = ResourceConfig.from_file(resource_config)

            if  rcs :
                for rc in rcs:
                    logger.info("Loaded resource configurations for %s" % rc)
                    self._resource_configs[rc] = rcs[rc].as_dict() 

        else :
            self._resource_configs [resource_config.name] = resource_config.as_dict()

    # -------------------------------------------------------------------------
    #
    def get_resource_config (self, resource_key):
        """Returns a dictionary of the requested resource config
        """

        if  resource_key in self._resource_aliases :
            logger.warning ("using alias '%s' for deprecated resource key '%s'" \
                         % (self._resource_aliases[resource_key], resource_key))
            resource_key = self._resource_aliases[resource_key]

        if  resource_key not in self._resource_configs:
            error_msg = "Resource key '%s' is not known." % resource_key
            raise PilotException(error_msg)

        return self._resource_configs[resource_key]




#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.database
   :platform: Unix
   :synopsis: Database functions.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os 
import saga
import datetime
import gridfs
import radical.utils as ru

from pymongo import *
from bson.objectid import ObjectId

from radical.pilot.states import *
from radical.pilot.utils  import DBConnectionInfo

COMMAND_CANCEL_PILOT        = "Cancel_Pilot"
COMMAND_CANCEL_COMPUTE_UNIT = "Cancel_Compute_Unit"
COMMAND_KEEP_ALIVE          = "Keep_Alive"
COMMAND_FIELD               = "commands"
COMMAND_TYPE                = "type"
COMMAND_ARG                 = "arg"
COMMAND_TIME                = "time"

# -----------------------------------------------------------------------------
#
class DBException(Exception):

    # -------------------------------------------------------------------------
    #
    def __init__(self, msg, obj=None):
        """Le constructeur. Creates a new exception object.
        """
        Exception.__init__(self, msg)
        self._obj = obj


# -----------------------------------------------------------------------------
#
class DBEntryExistsException(Exception):

    # -------------------------------------------------------------------------
    #
    def __init__(self, msg, obj=None):
        """Le constructeur. Creates a new exception object.
        """
        Exception.__init__(self, msg)
        self._obj = obj


#-----------------------------------------------------------------------------
#
class Session():

    #--------------------------------------------------------------------------
    #
    def __init__(self, db_url, db_name="radicalpilot"):
        """ Le constructeur. Should not be called directrly, but rather
            via the static methods new() or reconnect().
        """

        url = ru.Url (db_url)

        if  db_name :
            url.path = db_name

        mongo, db, dbname, pname, cname = ru.mongodb_connect (url)

        self._client = mongo
        self._db     = db
        self._dburl  = str(url)
        self._dbname = dbname
        self._dbauth = "%s:%s" % (url.username, url.password)

        self._session_id = None

        self._s  = None

        self._w  = None
        self._um = None

        self._p  = None
        self._pm = None

    #--------------------------------------------------------------------------
    #
    @staticmethod
    def new(sid, db_url, db_name="radicalpilot"):
        """ Creates a new session (factory method).
        """
        creation_time = datetime.datetime.utcnow()

        dbs = Session(db_url, db_name)
        dbs._create(sid, creation_time)

        connection_info = DBConnectionInfo(
            session_id=sid,
            dbname=dbs._dbname,
            dbauth=dbs._dbauth,
            dburl=dbs._dburl
        )

        return (dbs, creation_time, connection_info)

    #--------------------------------------------------------------------------
    #
    def _create(self, sid, creation_time):
        """ Creates a new session (private).

            A session is a distinct collection with three sub-collections
            in MongoDB:

            radical.pilot.<sid>    | Base collection. Holds some metadata.   | self._s
            radical.pilot.<sid>.cu | Collection holding all compute units.   | self._w
            radical.pilot.<sid>.um | Collection holding all unit managers.   | self._um
            radical.pilot.<sid>.p  | Collection holding all pilots.          | self._p
            radical.pilot.<sid>.pm | Collection holding all pilot managers.  | self._pm

            All collections are created with a new session. Since MongoDB
            uses lazy-create, they only appear in the database after the
            first insert. That's ok.
        """

        # make sure session doesn't exist already
        if  sid :
            if  self._db[sid].count() != 0 :
                raise DBEntryExistsException ("Session '%s' already exists." % sid)

        # remember session id
        self._session_id = sid

        self._s = self._db["%s" % sid]
        self._s.insert( 
            {
                "_id"  : ObjectId(sid),
                "created"          : creation_time,
                "last_reconnect"   : None
            }
        )

        # Create the collection shortcut:
        self._w  = self._db["%s.cu" % sid]
        self._um = self._db["%s.um" % sid] 

        self._p  = self._db["%s.p"  % sid]
        self._pm = self._db["%s.pm" % sid] 

    #--------------------------------------------------------------------------
    #
    @staticmethod
    def reconnect(sid, db_url, db_name="radical.pilot"):
        """ Reconnects to an existing session.

            Here we simply check if a radical.pilot.<sid> collection exists.
        """
        dbs = Session(db_url, db_name)
        session_info = dbs._reconnect(sid)

        connection_info = DBConnectionInfo(
            session_id=sid,
            dbname=dbs._dbname,
            dbauth=dbs._dbauth,
            dburl=dbs._dburl
        )

        return (dbs, session_info, connection_info)

    #--------------------------------------------------------------------------
    #
    def _reconnect(self, sid):
        """ Reconnects to an existing session (private).
        """
        # make sure session exists
        #if sid not in self._db.collection_names():
        #    raise DBEntryDoesntExistException("Session with id '%s' doesn't exists." % sid)

        self._s = self._db["%s" % sid]
        cursor = self._s.find({"_id": ObjectId(sid)})

        self._s.update({"_id"  : ObjectId(sid)},
                       {"$set" : {"last_reconnect" : datetime.datetime.utcnow()}}
        )

        cursor = self._s.find({"_id": ObjectId(sid)})

        # cursor -> dict
        #if len(cursor) != 1:
        #    raise DBEntryDoesntExistException("Session with id '%s' doesn't exists." % sid)

        self._session_id = sid

        # Create the collection shortcut:
        self._w  = self._db["%s.cu" % sid]
        self._um = self._db["%s.um" % sid]

        self._p  = self._db["%s.p"  % sid]
        self._pm = self._db["%s.pm" % sid]

        try:
            return cursor[0]
        except:
            raise Exception("Couldn't find Session UID '%s' in database." % sid)

    #--------------------------------------------------------------------------
    #
    @property
    def session_id(self):
        """ Returns the session id.
        """
        return self._session_id

    #--------------------------------------------------------------------------
    #
    def delete(self):
        """ Removes a session and all associated collections from the DB.
        """
        if self._s is None:
            raise DBException("No active session.")

        for collection in [self._s, self._w, self._um, self._p, self._pm]:
            collection.drop()
            collection = None

    #--------------------------------------------------------------------------
    #
    def insert_pilot_manager(self, pilot_manager_data, pilot_launcher_workers):
        """ Adds a pilot managers to the list of pilot managers.

            Pilot manager IDs are just kept for book-keeping.
        """
        if self._s is None:
            raise Exception("No active session.")

        pilot_manager_json = {"data": pilot_manager_data,
                              "pilot_launcher_workers": pilot_launcher_workers}
        result = self._pm.insert(pilot_manager_json)

        # return the object id as a string
        return str(result)

    #--------------------------------------------------------------------------
    #
    def list_pilot_manager_uids(self):
        """ Lists all pilot managers.
        """
        if self._s is None:
            raise Exception("No active session.")

        pilot_manager_uids = []
        cursor = self._pm.find()

        # cursor -> dict
        for obj in cursor:
            pilot_manager_uids.append(str(obj['_id']))
        return pilot_manager_uids

    #--------------------------------------------------------------------------
    #
    def get_compute_unit_stdout(self, unit_uid):
        """Returns the ComputeUnit's unit's stdout.
        """
        if self._s is None:
            raise Exception("No active session.")

        cursor = self._w.find({"_id": ObjectId(unit_uid)})

        return cursor[0]['stdout']

    #--------------------------------------------------------------------------
    #
    def get_compute_unit_stderr(self, unit_uid):
        """Returns the ComputeUnit's unit's stderr.
        """
        if self._s is None:
            raise Exception("No active session.")

        cursor = self._w.find({"_id": ObjectId(unit_uid)})

        return cursor[0]['stderr']

    #--------------------------------------------------------------------------
    #
    def update_pilot_state(self, pilot_uid, started=None, finished=None,
                           submitted=None, state=None, sagajobid=None,
                           sandbox=None, logs=None):

        """Updates the information of a pilot.
        """
        if self._s is None:
            raise Exception("No active session.")

        # construct the update query
        set_query = dict()
        push_query = dict()

        if state is not None:
            set_query["state"] = state
            push_query["statehistory"] = [{'state': state, 'timestamp': datetime.datetime.utcnow()}]

        if started is not None:
            set_query["started"] = started

        if finished is not None:
            set_query["finished"] = finished

        if submitted is not None:
            set_query["submitted"] = submitted

        if sagajobid is not None:
            set_query["sagajobid"] = sagajobid

        if sandbox is not None:
            set_query["sandbox"] = sandbox

        if logs is not None:
            push_query["log"] = logs

        # update pilot entry.
        self._p.update(
            {"_id": ObjectId(pilot_uid)},
            {"$set": set_query, "$pushAll": push_query},
            multi=True
        )

    #--------------------------------------------------------------------------
    #
    def insert_pilot(self, pilot_uid, pilot_manager_uid, pilot_description,
        sandbox):
        """Adds a new pilot document to the database.
        """
        if self._s is None:
            raise Exception("No active session.")

        ts = datetime.datetime.utcnow()

        pilot_doc = {
            "_id":            pilot_uid,
            "description":    pilot_description.as_dict(),
            "submitted":      datetime.datetime.utcnow(),
            "input_transfer_started": None,
            "input_transfer_finished": None,
            "started":        None,
            "finished":       None,
            "heartbeat":      None,
            "output_transfer_started": None,
            "output_transfer_finished": None,
            "nodes":          None,
            "cores_per_node": None,
            "sagajobid":      None,
            "sandbox":        sandbox,
            "state":          PENDING_LAUNCH,
            "statehistory":   [{"state": PENDING_LAUNCH, "timestamp": ts}],
            "log":            [],
            "pilotmanager":   pilot_manager_uid,
            "unitmanager":    None,
            "commands":       []
        }

        self._p.insert(pilot_doc, upsert=False)

        return str(pilot_uid), pilot_doc

    #--------------------------------------------------------------------------
    #
    def list_pilot_uids(self, pilot_manager_uid=None):
        """ Lists all pilots for a pilot manager.
        """
        if self._s is None:
            raise Exception("No active session.")

        pilot_ids = []

        if pilot_manager_uid is not None:
            cursor = self._p.find({"pilotmanager": pilot_manager_uid})
        else:
            cursor = self._p.find()

        # cursor -> dict
        for obj in cursor:
            pilot_ids.append(str(obj['_id']))
        return pilot_ids

    #--------------------------------------------------------------------------
    #
    def get_pilots(self, pilot_manager_id=None, pilot_ids=None):
        """ Get a pilot
        """
        if self._s is None:
            raise Exception("No active session.")

        if pilot_manager_id is None and pilot_ids is None:
            raise Exception(
                "pilot_manager_id and pilot_ids can't both be None.")

        if pilot_ids is None:
            cursor = self._p.find({"pilotmanager": pilot_manager_id})
        else:

            if not isinstance(pilot_ids, list):
                pilot_ids = [pilot_ids]

            # convert ids to object ids
            pilot_oid = []
            for pid in pilot_ids:
                pilot_oid.append(ObjectId(pid))
            cursor = self._p.find({"_id": {"$in": pilot_oid}})

        pilots_json = []
        for obj in cursor:
            pilots_json.append(obj)

        return pilots_json

    #--------------------------------------------------------------------------
    #
    def send_command_to_pilot(self, cmd, arg=None, pilot_manager_id=None, pilot_ids=None):
        """ Send a command to one or more pilots.
        """
        if self._s is None:
            raise Exception("No active session.")

        if pilot_manager_id is None and pilot_ids is None:
            raise Exception("Either Pilot Manager or Pilot needs to be specified.")

        if pilot_manager_id is not None and pilot_ids is not None:
            raise Exception("Pilot Manager and Pilot can not be both specified.")

        command = {COMMAND_FIELD: {COMMAND_TYPE: cmd,
                                   COMMAND_ARG:  arg,
                                   COMMAND_TIME: datetime.datetime.utcnow()
        }}

        if pilot_ids is None:
            # send command to all pilots that are known to the
            # pilot manager.
            self._p.update(
                {"pilotmanager": pilot_manager_id},
                {"$push": command},
                multi=True
            )
        else:
            if not isinstance(pilot_ids, list):
                pilot_ids = [pilot_ids]
            # send command to selected pilots if pilot_ids are
            # specified convert ids to object ids
            for pid in pilot_ids:
                self._p.update(
                    {"_id": ObjectId(pid)},
                    {"$push": command}
                )


    #--------------------------------------------------------------------------
    #
    def publish_compute_pilot_callback_history(self, pilot_uid, callback_history):

        if self._s is None:
            raise Exception("No active session.")

        self._p.update({"_id": ObjectId(pilot_uid)},
                       {"$set": {"callbackhistory": callback_history}})

    #--------------------------------------------------------------------------
    #
    def get_compute_units(self, unit_manager_id, unit_ids=None):
        """ Get yerself a bunch of compute units.
        """
        if self._s is None:
            raise Exception("No active session.")

        if unit_ids is None:
            cursor = self._w.find(
                {"unitmanager": unit_manager_id}
            )

        else:
            # convert ids to object ids
            unit_oid = []
            for wid in unit_ids:
                unit_oid.append(ObjectId(wid))

            cursor = self._w.find(
                {"_id": {"$in": unit_oid},
                 "unitmanager": unit_manager_id}
            )

        units_json = []
        for obj in cursor:
            units_json.append(obj)

        return units_json

    #--------------------------------------------------------------------------
    #
    def set_all_running_compute_units(self, pilot_id, state, log):
        """Update the state and the log of all compute units belonging to
           a specific pilot.
        """
        ts = datetime.datetime.utcnow()

        if self._s is None:
            raise Exception("No active session.")

        self._w.update({"pilot": pilot_id, "state": { "$in": [EXECUTING, PENDING_EXECUTION, SCHEDULING]}},
                       {"$set": {"state": state},
                        "$push": {"statehistory": {"state": state, "timestamp": ts},
                                  "log": log}
                       })

    #--------------------------------------------------------------------------
    #
    def set_compute_unit_state(self, unit_id, state, log):
        """Update the state and the log of one or more ComputeUnit(s).
        """
        ts = datetime.datetime.utcnow()

        if self._s is None:
            raise Exception("No active session.")

        self._w.update({"_id": ObjectId(unit_id)},
                       {"$set":     {"state": state},
                        "$push": {"statehistory": {"state": state, "timestamp": ts}},
                        "$pushAll": {"log": log}})

    #--------------------------------------------------------------------------
    #
    def get_compute_unit_states(self, unit_manager_id, unit_ids=None):
        """ Get yerself a bunch of compute units.
        """
        if self._s is None:
            raise Exception("No active session.")

        if unit_ids is None:
            cursor = self._w.find(
                {"unitmanager": unit_manager_id},
                {"state": 1}
            )

        else:
            # convert ids to object ids
            unit_oid = []
            for wid in unit_ids:
                unit_oid.append(ObjectId(wid))

            cursor = self._w.find(
                {"_id": {"$in": unit_oid},
                 "unitmanager": unit_manager_id},
                {"state": 1}
            )

        unit_states = []
        for obj in cursor:
            unit_states.append(obj['state'])

        return unit_states

    #--------------------------------------------------------------------------
    #
    def insert_unit_manager(self, scheduler, input_transfer_workers, output_transfer_workers):
        """ Adds a unit managers to the list of unit managers.

            Unit manager IDs are just kept for book-keeping.
        """
        if self._s is None:
            raise Exception("No active session.")

        result = self._um.insert(
            {"scheduler": scheduler,
             "input_transfer_workers": input_transfer_workers,
             "output_transfer_workers": output_transfer_workers }
        )

        # return the object id as a string
        return str(result)

    #--------------------------------------------------------------------------
    #
    def get_unit_manager(self, unit_manager_id):
        """ Get a unit manager.
        """
        if self._s is None:
            raise DBException("No active session.")

        cursor = self._um.find({"_id": ObjectId(unit_manager_id)})

        if cursor.count() != 1:
            msg = "No unit manager with id %s found in DB." % unit_manager_id
            raise DBException(msg=msg)

        try:
            return cursor[0]
        except:
            msg = "No UnitManager with id '%s' found in database." % unit_manager_id
            raise DBException(msg=msg)

    #--------------------------------------------------------------------------
    #
    def get_pilot_manager(self, pilot_manager_id):
        """ Get a unit manager.
        """
        if self._s is None:
            raise DBException("No active session.")

        cursor = self._pm.find({"_id": ObjectId(pilot_manager_id)})

        try:
            return cursor[0]
        except:
            msg = "No pilot manager with id '%s' found in DB." % pilot_manager_id
            raise DBException(msg=msg)


    #--------------------------------------------------------------------------
    #
    def list_unit_manager_uids(self):
        """ Lists all pilot managers.
        """
        if self._s is None:
            raise Exception("No active session.")

        unit_manager_uids = []
        cursor = self._um.find()

        # cursor -> dict
        for obj in cursor:
            unit_manager_uids.append(str(obj['_id']))
        return unit_manager_uids

    #--------------------------------------------------------------------------
    #
    def unit_manager_add_pilots(self, unit_manager_id, pilot_ids):
        """ Adds a pilot from a unit manager.
        """
        if self._s is None:
            raise Exception("No active session.")

        for pilot_id in pilot_ids:
            self._p.update({"_id": ObjectId(pilot_id)},
                           {"$set": {"unitmanager": unit_manager_id}},
                           True)

    #--------------------------------------------------------------------------
    #
    def unit_manager_remove_pilots(self, unit_manager_id, pilot_ids):
        """ Removes one or more pilots from a unit manager.
        """
        if self._s is None:
            raise Exception("No active session.")

        # Add the ids to the pilot's queue
        for pilot_id in pilot_ids:
            self._p.update({"_id": ObjectId(pilot_id)},
                           {"$set": {"unitmanager": None}}, True)

    #--------------------------------------------------------------------------
    #
    def unit_manager_list_pilots(self, unit_manager_uid):
        """ Lists all pilots associated with a unit manager.
        """
        if self._s is None:
            raise Exception("No active session.")

        cursor = self._p.find({"unitmanager": unit_manager_uid})

        # cursor -> dict
        pilot_ids = []
        for obj in cursor:
            pilot_ids.append(str(obj['_id']))
        return pilot_ids

    #--------------------------------------------------------------------------
    #
    def unit_manager_list_compute_units(self, unit_manager_uid):
        """ Lists all compute units associated with a unit manager.
        """
        if self._s is None:
            raise Exception("No active session.")

        cursor = self._w.find({"unitmanager": unit_manager_uid})

        # cursor -> dict
        unit_ids = []
        for obj in cursor:
            unit_ids.append(str(obj['_id']))
        return unit_ids

    #--------------------------------------------------------------------------
    #
    def assign_compute_units_to_pilot(self, units, pilot_uid, pilot_sandbox):
        """Assigns one or more compute units to a pilot.
        """

        if  not units :
            return

        if  self._s is None:
            raise Exception("No active session.")

        # Make sure we work on a list.
        if not isinstance(units, list):
            units = [units]

        bulk = self._w.initialize_ordered_bulk_op ()

        for unit in units :

            bulk.find   ({"_id" : ObjectId(unit.uid)}) \
                .update ({"$set": {"description"   : unit.description.as_dict(),
                                   "pilot"         : pilot_uid,
                                   "pilot_sandbox" : pilot_sandbox,
                                   "sandbox"       : unit.sandbox,
                                   "FTW_Input_Status": unit.FTW_Input_Status,
                                   "FTW_Input_Directives": unit.FTW_Input_Directives,
                                   "Agent_Input_Status": unit.Agent_Input_Status,
                                   "Agent_Input_Directives": unit.Agent_Input_Directives,
                                   "FTW_Output_Status": unit.FTW_Output_Status,
                                   "FTW_Output_Directives": unit.FTW_Output_Directives,
                                   "Agent_Output_Status": unit.Agent_Output_Status,
                                   "Agent_Output_Directives": unit.Agent_Output_Directives
                        }})
        result = bulk.execute()

        # TODO: log result.
        # WHY DON'T WE HAVE A LOGGER HERE?


    #--------------------------------------------------------------------------
    #
    def publish_compute_unit_callback_history(self, unit_uid, callback_history):

        if self._s is None:
            raise Exception("No active session.")

        self._w.update({"_id": ObjectId(unit_uid)},
                       {"$set": {"callbackhistory": callback_history}})

    #--------------------------------------------------------------------------
    #
    def insert_compute_units(self, unit_manager_uid, units, unit_log):
        """ Adds one or more compute units to the database and sets their state
            to 'PENDING'.
        """
        if self._s is None:
            raise Exception("No active session.")

        # Make sure we work on a list.
        if not isinstance(units, list):
            units = [units]

        unit_docs = list()
        results = dict()

        for unit in units:

            ts = datetime.datetime.utcnow()

            unit_json = {
                "_id":           ObjectId(unit.uid),
                "description":   unit.description.as_dict(),
                "unitmanager":   unit_manager_uid,
                "pilot":         None,
                "pilot_sandbox": None,
                "state":         unit._local_state,
                "statehistory":  [{"state": unit._local_state, "timestamp": ts}],
                "submitted":     datetime.datetime.utcnow(),
                "started":       None,
                "finished":      None,
                "exec_locs":     None,
                "exit_code":     None,
                #"workdir":       "unit-"+unit.uid,
                "sandbox":       None,
                "stdout":        None,
                "stderr":        None,
                "log":           unit_log,
                "FTW_Input_Status": None,
                "FTW_Input_Directives": None,
                "Agent_Input_Status": None,
                "Agent_Input_Directives": None,
                "FTW_Output_Status": None,
                "FTW_Output_Directives": None,
                "Agent_Output_Status": None,
                "Agent_Output_Directives": None
            }
            unit_docs.append(unit_json)
            results[unit.uid] = unit_json

        unit_uids = self._w.insert(unit_docs)

        assert len(unit_docs) == len(unit_uids)
        assert len(results) == len(unit_uids)

        return results



#!/usr/bin/env python

"""
.. module:: radical.pilot.agent
   :platform: Unix
   :synopsis: An agent "skeleton" for RADICAL-Pilot. 

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import ast
import sys
import time
import errno
import pipes
import Queue
import signal
import gridfs
import pymongo
import optparse
import logging
import datetime
import hostlist
import traceback
import threading 
import subprocess
import multiprocessing

from bson.objectid import ObjectId

#--------------------------------------------------------------------------
# Configure the logger
LOGGER = logging.getLogger('radical.pilot.agent')
LOGGER.setLevel(logging.INFO)
ch = logging.FileHandler("AGENT.LOG")
ch.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
LOGGER.addHandler(ch)

#---------------------------------------------------------------------------
#
def pilot_FAILED(mongodb_handle, pilot_uid, message):
    """Updates the state of one or more pilots.
    """
    pilot_collection = mongo_db["%s.p"  % options.session_id]

    LOGGER.error(message)      
    ts = datetime.datetime.utcnow()

    pilot_collection.update({"_id": ObjectId(pilot_uid)}, 
        {"$push": {"log" : message,
                   "statehistory": {"state": 'Failed', "timestamp": ts}},
         "$set":  {"state": 'Failed',
                   "finished": ts}

        })

#---------------------------------------------------------------------------
#
def pilot_CANCELED(mongodb_handle, pilot_uid, message):
    """Updates the state of one or more pilots.
    """
    pilot_collection = mongo_db["%s.p"  % options.session_id]

    LOGGER.warning(message)
    ts = datetime.datetime.utcnow()

    pilot_collection.update({"_id": ObjectId(pilot_uid)}, 
        {"$push": {"log" : message,
                   "statehistory": {"state": 'Canceled', "timestamp": ts}},
         "$set":  {"state": 'Canceled',
                   "finished": ts}
        })

#---------------------------------------------------------------------------
#
def pilot_DONE(mongodb_handle, pilot_uid, message):
    """Updates the state of one or more pilots.
    """
    pilot_collection = mongo_db["%s.p"  % options.session_id]

    LOGGER.info(message)
    ts = datetime.datetime.utcnow()

    pilot_collection.update({"_id": ObjectId(pilot_uid)}, 
        {"$push": {"log" : message,
                   "statehistory": {"state": 'Done', "timestamp": ts}},
         "$set": {"state": 'Done',
                  "finished": ts}
        })


# ----------------------------------------------------------------------------
#
class Agent(threading.Thread):

    # ------------------------------------------------------------------------
    #
    def __init__(self, workdir, runtime, mongodb_url, mongodb_name, pilot_id, session_id):
        """Le Constructeur creates a new Agent instance.
        """
        threading.Thread.__init__(self)
        self.daemon      = True
        self.lock        = threading.Lock()
        self._terminate  = threading.Event()

        self._workdir    = workdir
        self._pilot_id   = pilot_id

        self._runtime    = runtime
        self._starttime  = None

        mongo_client = pymongo.MongoClient(mongodb_url)
        self.mongo_db = mongo_client[mongodb_name]
        self.pilot_collection = self.mongo_db["%s.p"  % session_id]
        self.computeunit_collection = self.mongo_db["%s.cu"  % session_id]

    # ------------------------------------------------------------------------
    #
    def stop(self):
        """Terminate the agent main loop.
        """
        # Next, we set our own termination signal
        self._terminate.set()

    # ------------------------------------------------------------------------
    #
    def run(self):
        """Starts the thread when Thread.start() is called.
        """
        # first order of business: set the start time and state of the pilot
        LOGGER.info("Agent started. Database updated.")
        ts = datetime.datetime.utcnow()

        # ---------------------------------
        # Update the pilot's database entry
        self.pilot_collection.update(
            {"_id": ObjectId(self._pilot_id)}, 
            {"$set": {"state"          : "Running",
                      "nodes"          : "SKEL-AGENT-None",
                      "cores_per_node" : "SKEL-AGENT-None",
                      "started"        : ts},
             "$push": {"statehistory": {"state": 'Running', "timestamp": ts}}
            })

        self._starttime = time.time()

        # ---------------------------------
        # This is the main thread loop.
        while True:
            try:
                # Exit the main loop if terminate is set. 
                if self._terminate.isSet():
                    pilot_CANCELED(self.mongo_db, self._pilot_id, "Agent terminated.")
                    break

                # Make sure that we haven't exceeded the agent runtime. if 
                # we have, terminate. 
                if time.time() >= self._starttime + (int(self._runtime) * 60):
                    runtime_limit = str(int(self._runtime)*60)
                    message = "Agent has reached its runtime limit (%s seconds)." % runtime_limit
                    pilot_DONE(self.mongo_db, self._pilot_id, message)
                    break

                # Check if there's a command waiting.
                cursor = self.pilot_collection.find({"_id": ObjectId(self._pilot_id)})
                command = cursor[0]['command']
                if command is not None:
                    if command.lower() == "cancel":
                        pilot_CANCELED(self.mongo_db, self._pilot_id, "Command 'CANCEL' received. Terminating.")
                        break
                    else:
                        LOGGER.warning("Received unknown command '%s'." % command )


                # Check if there are compute units waiting for execution
                ts = datetime.datetime.utcnow()

                computeunits = self.computeunit_collection.find_and_modify(
                query={"pilot" : self._pilot_id,
                       "state" : "PendingExecution"},
                update={"$set" : {"state": "Running"},
                "$push": {"statehistory": {"state": "RunningX", "timestamp": ts}}}
                #limit=BULK_LIMIT
                )

                # There are new compute units in the cu_queue on the database.
                # Get the corresponding cu entries
                if computeunits is not None:
                    if not isinstance(computeunits, list):
                        computeunits = [computeunits]

                    for cu in computeunits:
                        LOGGER.info("Processing ComputeUnit: %s" % cu)

                        # Create the task working directory if it doesn't exist
                        cu_workdir = "%s/unit-%s" % (self._workdir, str(cu["_id"]))
                        if not os.path.exists(cu_workdir):
                            os.makedirs(cu_workdir)

                        # Create bogus STDOUT and STDERR
                        open("%s/STDOUT" % cu_workdir, 'a').close()
                        open("%s/STDERR" % cu_workdir, 'a').close()

                        if cu['description']['output_data'] is not None:
                            state = "PendingOutputTransfer"
                        else:
                            state = "Done"

                        self.computeunit_collection.update({"_id": cu["_id"]}, 
                            {"$set": {"state"         : state,
                                      "started"       : "SKEL-AGENT-None",
                                      "finished"      : "SKEL-AGENT-None",
                                      "exec_locs"     : "SKEL-AGENT-None",
                                      "exit_code"     : "SKEL-AGENT-None",
                                      "stdout_id"     : "SKEL-AGENT-None",
                                      "stderr_id"     : "SKEL-AGENT-None"},
                             "$push": {"statehistory": {"state": state, "timestamp": ts}}
                            })

            except Exception, ex:
                # If we arrive here, there was an exception in the main loop.
                pilot_FAILED(self.mongo_db, self._pilot_id, 
                    "ERROR in agent main loop: %s. %s" % (str(ex), traceback.format_exc()))
                return 

        # MAIN LOOP TERMINATED
        pilot_DONE(self.mongo_db, self._pilot_id, "Pilot main loop completed.")
        return



# ================================================================================
# ================================================================================
#
# BELOW THIS LINE NOTHING NEEDS TO BE TOUCHED (I THINK). IT'S MOSTLY SIGNAL 
# HANDLING STUFF, COMMAND-LINE PARSING, ETC.

def parse_commandline():

    parser = optparse.OptionParser()

    parser.add_option('-d', '--mongodb-url',
                      metavar='URL',
                      dest='mongodb_url',
                      help='Specifies the MongoDB Url.')

    parser.add_option('-n', '--database-name',
                      metavar='URL',
                      dest='database_name',
                      help='Specifies the MongoDB database name.')

    parser.add_option('-s', '--session-id',
                      metavar='SID',
                      dest='session_id',
                      help='Specifies the Session ID.')

    parser.add_option('-p', '--pilot-id',
                      metavar='PID',
                      dest='pilot_id',
                      help='Specifies the Pilot ID.')

    parser.add_option('-w', '--workdir',
                      metavar='DIRECTORY',
                      dest='workdir',
                      help='Specifies the base (working) directory for the agent. [default: %default]',
                      default='.')

    parser.add_option('-c', '--cores',
                      metavar='CORES',
                      dest='cores',
                      help='Specifies the number of cores to allocate.')

    parser.add_option('-t', '--runtime',
                      metavar='RUNTIME',
                      dest='runtime',
                      help='Specifies the agent runtime in minutes.')

    parser.add_option('-l', '--launch-method', 
                      metavar='METHOD',
                      dest='launch_method',
                      help='Enforce a specific launch method (AUTO, LOCAL, SSH, MPIRUN, APRUN). [default: %default]',
                      default="AUTO")

    parser.add_option('-V', '--version', 
                      metavar='VERSION ',
                      dest='package_version',
                      help='The RADICAL-Pilot package version.')

    # parse the whole shebang
    (options, args) = parser.parse_args()

    if options.mongodb_url is None:
        parser.error("You must define MongoDB URL (-d/--mongodb-url). Try --help for help.")
    elif options.database_name is None:
        parser.error("You must define a database name (-n/--database-name). Try --help for help.")
    elif options.session_id is None:
        parser.error("You must define a session id (-s/--session-id). Try --help for help.")
    elif options.pilot_id is None:
        parser.error("You must define a pilot id (-p/--pilot-id). Try --help for help.")
    elif options.cores is None:
        parser.error("You must define the number of cores (-c/--cores). Try --help for help.")
    elif options.runtime is None:
        parser.error("You must define the agent runtime (-t/--runtime). Try --help for help.")
    elif options.package_version is None:
        parser.error("You must pass the RADICAL-Pilot package version (-v/--version). Try --help for help.")


    #if options.launch_method is not None: 
    #    valid_options = [LAUNCH_METHOD_AUTO, LAUNCH_METHOD_LOCAL, LAUNCH_METHOD_SSH, LAUNCH_METHOD_MPIRUN, LAUNCH_METHOD_APRUN]
    #    if options.launch_method.upper() not in valid_options:
    #        parser.error("--launch-method must be one of these: %s" % valid_options)

    return options

#-----------------------------------------------------------------------------
#
if __name__ == "__main__":

    # parse command line options
    options = parse_commandline()

    LOGGER.info("RADICAL-Pilot agent (radical-pilot-agent-skeleton.py) for package/API version %s" % options.package_version)


    #--------------------------------------------------------------------------
    # Establish database connection
    try:
        mongo_client = pymongo.MongoClient(options.mongodb_url)
        mongo_db     = mongo_client[options.database_name]

    except Exception, ex:
        LOGGER.error("Couldn't establish database connection: %s" % str(ex))
        sys.exit(1)

    #--------------------------------------------------------------------------
    # Some singal handling magic 
    def sigint_handler(signal, frame):
        msg = 'Caught SIGINT. EXITING.'
        pilot_CANCELED(mongo_db, options.pilot_id, msg)
        sys.exit(0)
    signal.signal(signal.SIGINT, sigint_handler)

    def sigalarm_handler(signal, frame):
        msg = 'Caught SIGALRM (Walltime limit reached?). EXITING'
        pilot_CANCELED(mongo_db, options.pilot_id, msg)
        sys.exit(0)
    signal.signal(signal.SIGALRM, sigalarm_handler)

    #--------------------------------------------------------------------------
    # Launch the agent thread
    try:
        if options.workdir is '.':
            workdir = os.getcwd()
        else:
            workdir = options.workdir

        agent = Agent(workdir=workdir,
                      runtime=options.runtime,
                      mongodb_url=options.mongodb_url,
                      mongodb_name=options.database_name,
                      pilot_id=options.pilot_id,
                      session_id=options.session_id)

        agent.start()
        agent.join()
        sys.exit(0)

    except Exception, ex:
        msg = "Error during agent execution: %s" % str(ex)
        LOGGER.error(msg)
        pilot_FAILED(mongo_db, options.pilot_id, msg)
        agent.stop()
        sys.exit(1)



#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.pilot_manager
   :platform: Unix
   :synopsis: Provides the interface for the PilotManager class.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import time
import glob
import copy

from radical.pilot.states import *
from radical.pilot.exceptions import *

from radical.pilot.object import Object
from radical.pilot.controller import PilotManagerController
from radical.pilot.compute_pilot import ComputePilot
from radical.pilot.utils.logger import logger
from radical.pilot.exceptions import PilotException, BadParameter
from radical.pilot.resource_config import ResourceConfig

# -----------------------------------------------------------------------------
#
class PilotManager(Object):
    """A PilotManager holds :class:`radical.pilot.ComputePilot` instances that are
    submitted via the :func:`radical.pilot.PilotManager.submit_pilots` method.

    It is possible to attach one or more :ref:`chapter_machconf`
    to a PilotManager to outsource machine specific configuration
    parameters to an external configuration file.

    Each PilotManager has a unique identifier :data:`radical.pilot.PilotManager.uid`
    that can be used to re-connect to previoulsy created PilotManager in a
    given :class:`radical.pilot.Session`.

    **Example**::

        s = radical.pilot.Session(database_url=dbURL)

        pm1 = radical.pilot.PilotManager(session=s, resource_configurations=RESCONF)
        # Re-connect via the 'get()' method.
        pm2 = radical.pilot.PilotManager.get(session=s, pilot_manager_id=pm1.uid)

        # pm1 and pm2 are pointing to the same PilotManager
        assert pm1.uid == pm2.uid
    """

    # -------------------------------------------------------------------------
    #
    def __init__(self, session, pilot_launcher_workers=1, _reconnect=False):
        """Creates a new PilotManager and attaches is to the session.

        .. note:: The `resource_configurations` (see :ref:`chapter_machconf`)
                  parameter is currently mandatory for creating a new
                  PilotManager instance.

        **Arguments:**

            * **session** [:class:`radical.pilot.Session`]:
              The session instance to use.

            * **resource_configurations** [`string` or `list of strings`]:
              A list of URLs pointing to :ref:`chapter_machconf`. Currently
              `file://`, `http://` and `https://` URLs are supported.

              If one or more resource_configurations are provided, Pilots
              submitted  via this PilotManager can access the configuration
              entries in the  files via the :class:`ComputePilotDescription`.
              For example::

                  pm = radical.pilot.PilotManager(session=s)

                  pd = radical.pilot.ComputePilotDescription()
                  pd.resource = "futuregrid.india"  # defined in futuregrid.json
                  pd.cores    = 16
                  pd.runtime  = 5 # minutes

                  pilot = pm.submit_pilots(pd)

            * pilot_launcher_workers (`int`): The number of pilot launcher 
              worker processes to start in the background. 

        .. note:: `pilot_launcher_workers` can be used to tune RADICAL-Pilot's 
                  performance. However, you should only change the default values 
                  if you know what you are doing.

        **Returns:**

            * A new `PilotManager` object [:class:`radical.pilot.PilotManager`].

        **Raises:**
            * :class:`radical.pilot.PilotException`
        """
        self._session = session
        self._worker = None
        self._uid = None

        if _reconnect == True:
            return

        ###############################
        # Create a new pilot manager. #
        ###############################

        # Start a worker process fo this PilotManager instance. The worker
        # process encapsulates database access, persitency et al.
        self._worker = PilotManagerController(
            pilot_manager_uid=None,
            pilot_manager_data={},
            pilot_launcher_workers=pilot_launcher_workers, 
            session=self._session,
            db_connection=session._dbs,
            db_connection_info=session._connection_info)
        self._worker.start()

        self._uid = self._worker.pilot_manager_uid

        # Each pilot manager has a worker thread associated with it. The task
        # of the worker thread is to check and update the state of pilots, fire
        # callbacks and so on.
        self._session._pilot_manager_objects.append(self)
        self._session._process_registry.register(self._uid, self._worker)

    #--------------------------------------------------------------------------
    #
    def close(self, terminate=True):
        """Shuts down the PilotManager and its background workers in a 
        coordinated fashion.

        **Arguments:**

            * **terminate** [`bool`]: If set to True, all active pilots will 
              get canceled (default: False).

        """
        # Spit out a warning in case the object was already closed.
        if not self._uid:
            logger.warning("PilotManager object already closed.")
            return

        # If terminate is set, we cancel all pilots. 
        if terminate is True:
            self.cancel_pilots()

        # Shut down all worker processes if still active. stop() returns
        # only after a successful join(). 
        if self._worker is not None:
            # Stop the worker process
            self._worker.stop()
            # Remove worker from registry
            self._session._process_registry.remove(self._uid)

        logger.info("Closed PilotManager %s." % str(self._uid))
        self._uid = None

    #--------------------------------------------------------------------------
    #
    @classmethod
    def _reconnect(cls, session, pilot_manager_id):
        """PRIVATE: reconnect to an existing pilot manager.
        """
        uid_exists = PilotManagerController.uid_exists(
            db_connection=session._dbs,
            pilot_manager_uid=pilot_manager_id
        )

        if not uid_exists:
            raise BadParameter(
                "PilotManager with id '%s' not in database." % pilot_manager_id)

        obj = cls(session=session, _reconnect=True)
        obj._uid = pilot_manager_id

        # Retrieve or start a worker process fo this PilotManager instance.
        worker = session._process_registry.retrieve(pilot_manager_id)
        if worker is not None:
            obj._worker = worker
        else:
            obj._worker = PilotManagerController(
                pilot_manager_uid=pilot_manager_id,
                pilot_manager_data={},
                session=session,
                db_connection=session._dbs,
                db_connection_info=session._connection_info)
            session._process_registry.register(pilot_manager_id, obj._worker)

        # start the worker if it's not already running
        if obj._worker.is_alive() is False:
            obj._worker.start()

        return obj

    # -------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns a Python dictionary representation of the object.
        """
        self._assert_obj_is_valid()

        object_dict = {
            'uid': self.uid
        }
        return object_dict

    # -------------------------------------------------------------------------
    #
    def __str__(self):
        """Returns a string representation of the object.
        """
        return str(self.as_dict())

    # -------------------------------------------------------------------------
    #
    def submit_pilots(self, pilot_descriptions):
        """Submits a new :class:`radical.pilot.ComputePilot` to a resource.

        **Returns:**

            * One or more :class:`radical.pilot.ComputePilot` instances
              [`list of :class:`radical.pilot.ComputePilot`].

        **Raises:**

            * :class:`radical.pilot.PilotException`
        """
        # Check if the object instance is still valid.
        self._assert_obj_is_valid()

        # Implicit list conversion.
        return_list_type = True
        if  not isinstance(pilot_descriptions, list):
            return_list_type   = False
            pilot_descriptions = [pilot_descriptions]

        # Itereate over the pilot descriptions, try to create a pilot for
        # each one and append it to 'pilot_obj_list'.
        pilot_obj_list = list()

        for pilot_description in pilot_descriptions:

            if pilot_description.resource is None:
                error_msg = "ComputePilotDescription does not define mandatory attribute 'resource'."
                raise BadParameter(error_msg)

            elif pilot_description.runtime is None:
                error_msg = "ComputePilotDescription does not define mandatory attribute 'runtime'."
                raise BadParameter(error_msg)

            elif pilot_description.cores is None:
                error_msg = "ComputePilotDescription does not define mandatory attribute 'cores'."
                raise BadParameter(error_msg)

            resource_key = pilot_description.resource
            resource_cfg = self._session.get_resource_config(resource_key)

            # Check resource-specific mandatory attributes
            if "mandatory_args" in resource_cfg:
                for ma in resource_cfg["mandatory_args"]:
                    if getattr(pilot_description, ma) is None:
                        error_msg = "ComputePilotDescription does not define attribute '{0}' which is required for '{1}'.".format(ma, resource_key)
                        raise BadParameter(error_msg)


            # we expand and exchange keys in the resource config, depending on
            # the selected schema so better use a deep copy...
            import copy
            resource_cfg  = copy.deepcopy (resource_cfg)
            schema        = pilot_description['access_schema']

            if  not schema :
                if 'schemas' in resource_cfg :
                    schema = resource_cfg['schemas'][0]

            if  not schema in resource_cfg :
                import pprint
                pprint.pprint (resource_cfg)
                logger.warning ("schema %s unknown for resource %s -- continue with defaults" \
                             % (schema, resource_key))

            else :
                for key in resource_cfg[schema] :
                    # merge schema specific resource keys into the
                    # resource config
                    resource_cfg[key] = resource_cfg[schema][key]

            # If 'default_sandbox' is defined, set it.
            if pilot_description.sandbox is not None:
                if "valid_roots" in resource_cfg and resource_cfg["valid_roots"] is not None:
                    is_valid = False
                    for vr in resource_cfg["valid_roots"]:
                        if pilot_description.sandbox.startswith(vr):
                            is_valid = True
                    if is_valid is False:
                        raise BadParameter("Working directory for resource '%s' defined as '%s' but needs to be rooted in %s " % (resource_key, pilot_description.sandbox, resource_cfg["valid_roots"]))

            # After the sanity checks have passed, we can register a pilot
            # startup request with the worker process and create a facade
            # object.

            pilot = ComputePilot._create(
                pilot_description=pilot_description,
                pilot_manager_obj=self)

            pilot_uid = self._worker.register_start_pilot_request(
                pilot=pilot,
                resource_config=resource_cfg)

            pilot._uid = pilot_uid

            pilot_obj_list.append(pilot)

        # Implicit return value conversion
        if  return_list_type :
            return pilot_obj_list
        else:
            return pilot_obj_list[0]

    # -------------------------------------------------------------------------
    #
    def list_pilots(self):
        """Lists the unique identifiers of all :class:`radical.pilot.ComputePilot`
        instances associated with this PilotManager

        **Returns:**

            * A list of :class:`radical.pilot.ComputePilot` uids [`string`].

        **Raises:**

            * :class:`radical.pilot.PilotException`
        """
        # Check if the object instance is still valid.
        self._assert_obj_is_valid()

        # Get the pilot list from the worker
        return self._worker.list_pilots()

    # -------------------------------------------------------------------------
    #
    def get_pilots(self, pilot_ids=None):
        """Returns one or more :class:`radical.pilot.ComputePilot` instances.

        **Arguments:**

            * **pilot_uids** [`list of strings`]: If pilot_uids is set,
              only the Pilots with  the specified uids are returned. If
              pilot_uids is `None`, all Pilots are returned.

        **Returns:**

            * A list of :class:`radical.pilot.ComputePilot` objects
              [`list of :class:`radical.pilot.ComputePilot`].

        **Raises:**

            * :class:`radical.pilot.PilotException`
        """
        self._assert_obj_is_valid()

        
        return_list_type = True
        if (not isinstance(pilot_ids, list)) and (pilot_ids is not None):
            return_list_type = False
            pilot_ids = [pilot_ids]

        pilots = ComputePilot._get(pilot_ids=pilot_ids, pilot_manager_obj=self)

        if  return_list_type :
            return pilots
        else :
            return pilots[0]

    # -------------------------------------------------------------------------
    #
    def wait_pilots(self, pilot_ids=None,
                    state=[DONE, FAILED, CANCELED],
                    timeout=None):
        """Returns when one or more :class:`radical.pilot.ComputePilots` reach a
        specific state or when an optional timeout is reached.

        If `pilot_uids` is `None`, `wait_pilots` returns when **all** Pilots
        reach the state defined in `state`.

        **Arguments:**

            * **pilot_uids** [`string` or `list of strings`]
              If pilot_uids is set, only the Pilots with the specified uids are
              considered. If pilot_uids is `None` (default), all Pilots are
              considered.

            * **state** [`list of strings`]
              The state(s) that Pilots have to reach in order for the call
              to return.

              By default `wait_pilots` waits for the Pilots to reach
              a **terminal** state, which can be one of the following:

              * :data:`radical.pilot.DONE`
              * :data:`radical.pilot.FAILED`
              * :data:`radical.pilot.CANCELED`

            * **timeout** [`float`]
              Optional timeout in seconds before the call returns regardless
              whether the Pilots have reached the desired state or not.
              The default value **-1.0** never times out.

        **Raises:**

            * :class:`radical.pilot.PilotException`
        """
        self._assert_obj_is_valid()

        if not isinstance(state, list):
            state = [state]

        return_list_type = True
        if (not isinstance(pilot_ids, list)) and (pilot_ids is not None):
            return_list_type = False
            pilot_ids = [pilot_ids]


        start  = time.time()
        all_ok = False
        states = list()

        while not all_ok :

            pilots = self._worker.get_compute_pilot_data(pilot_ids=pilot_ids)
            all_ok = True
            states = list()

            for pilot in pilots :
                if  pilot['state'] not in state :
                    all_ok = False

                states.append (pilot['state'])

            # check timeout
            if  (None != timeout) and (timeout <= (time.time() - start)):
                if  not all_ok :
                    logger.debug ("wait timed out: %s" % states)
                break

            # sleep a little if this cycle was idle
            if  not all_ok :
                time.sleep (0.1)

        # done waiting
        if  return_list_type :
            return states
        else :
            return states[0]


    # -------------------------------------------------------------------------
    #
    def cancel_pilots(self, pilot_ids=None):
        """Cancels one or more ComputePilots.

        **Arguments:**

            * **pilot_uids** [`string` or `list of strings`]
              If pilot_uids is set, only the Pilots with the specified uids are
              canceled. If pilot_uids is `None`, all Pilots are canceled.

        **Raises:**

            * :class:`radical.pilot.PilotException`
        """
        # Check if the object instance is still valid.
        self._assert_obj_is_valid()

        # Implicit list conversion.
        if (not isinstance(pilot_ids, list)) and (pilot_ids is not None):
            pilot_ids = [pilot_ids]

        # Register the cancelation request with the worker.
        self._worker.register_cancel_pilots_request(pilot_ids=pilot_ids)


    # -------------------------------------------------------------------------
    #
    def register_callback(self, callback_function):
        """Registers a new callback function with the PilotManager.
        Manager-level callbacks get called if any of the ComputePilots managed
        by the PilotManager change their state.

        All callback functions need to have the same signature::

            def callback_func(obj, state)

        where ``object`` is a handle to the object that triggered the callback
        and ``state`` is the new state of that object.
        """
        self._assert_obj_is_valid()

        self._worker.register_manager_callback(callback_function)



""" (Compute) Unit tests
"""
import os
import sys
import radical.pilot
import unittest

import uuid
from copy import deepcopy
from radical.pilot.db import Session
from pymongo import MongoClient

# DBURL defines the MongoDB server URL and has the format mongodb://host:port.
# For the installation of a MongoDB server, refer to the MongoDB website:
# http://docs.mongodb.org/manual/installation/
DBURL = os.getenv("RADICAL_PILOT_DBURL")
if DBURL is None:
    print "ERROR: RADICAL_PILOT_DBURL (MongoDB server URL) is not defined."
    sys.exit(1)
    
DBNAME = os.getenv("RADICAL_PILOT_TEST_DBNAME")
if DBNAME is None:
    print "ERROR: RADICAL_PILOT_TEST_DBNAME (MongoDB database name) is not defined."
    sys.exit(1)


#-----------------------------------------------------------------------------
#
class TestIssue169(unittest.TestCase):
    # silence deprecation warnings under py3

    def setUp(self):
        # clean up fragments from previous tests
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def tearDown(self):
        # clean up after ourselves 
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def failUnless(self, expr):
        # St00pid speling.
        return self.assertTrue(expr)

    def failIf(self, expr):
        # St00pid speling.
        return self.assertFalse(expr)

    #-------------------------------------------------------------------------
    #
    def test__issue_169_part_1(self):
        """ https://github.com/radical-cybertools/radical.pilot/issues/169
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)

        pm = radical.pilot.PilotManager(session=session)

        cpd = radical.pilot.ComputePilotDescription()
        cpd.resource = "local.localhost"
        cpd.cores = 1
        cpd.runtime = 1
        cpd.sandbox = "/non-/existing/directory..."
        cpd.cleanup = True

        for i in range(0, 8):
            pilot = pm.submit_pilots(pilot_descriptions=cpd)
            pilot.wait(state=radical.pilot.FAILED, timeout=5*60)
            assert pilot.state == radical.pilot.FAILED, "State is %s instead of 'Failed'." % pilot.state

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__issue_169_part_2(self):
        """ https://github.com/radical-cybertools/radical.pilot/issues/169
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)

        pmgr = radical.pilot.PilotManager(session=session)
        
        cpd1 = radical.pilot.ComputePilotDescription()
        cpd1.resource = "local.localhost"
        cpd1.cores    = 1
        cpd1.runtime  = 1
        cpd1.sandbox  = "/tmp/radical.pilot.sandbox.unittests"
        cpd1.cleanup  = True

        cpd2 = radical.pilot.ComputePilotDescription()
        cpd2.resource = "local.localhost"
        cpd2.cores    = 1
        cpd2.runtime  = 1
        cpd2.sandbox  = "/tmp/radical.pilot.sandbox.unittests"
        cpd2.cleanup  = True

        pilots = pmgr.submit_pilots([cpd1, cpd2])

        pmgr.wait_pilots(timeout=10*60)
        
        for pilot in pilots:
            try :
                assert pilot.state == radical.pilot.DONE, "state: %s" % pilot.state
                assert pilot.stop_time  is not None,      "time : %s" % pilot.stop_time
                assert pilot.start_time is not None,      "time : %s" % pilot.start_time
            except :
                print 'pilot: %s (%s)' % (pilot.uid, pilot.state)
                for entry in pilot.state_history :
                    print '       %s : %s' % (entry.timestamp, entry.state)
                print '     : %s' % str(pilot.log)
                raise

        session.close()




#pylint: disable=C0301, C0103, W0212, R0903, W0232

"""
.. module:: radical.pilot.url
   :platform: Unix
   :synopsis: Implementation of the Url class (forwarded from SAGA-Python).

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

from saga import Url as sagaURL


# ------------------------------------------------------------------------------
#
class Url(sagaURL): 
    """Defines a URL class.
    """
    pass



#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.scheduler.RoundRobinScheduler
   :platform: Unix
   :synopsis: A multi-pilot, round-robin scheduler.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os 

from radical.pilot.utils.logger        import logger
from radical.pilot.scheduler.interface import Scheduler 

# -----------------------------------------------------------------------------
# 
class RoundRobinScheduler(Scheduler):
    """RoundRobinScheduler implements a multi-pilot, round-robin 
    scheduling algorithm.
    """

    # -------------------------------------------------------------------------
    #
    def __init__(self, manager, session):
        """Le constructeur.
        """

        self.manager = manager
        self.session = session
        self.pilots  = dict()
        self.idx     = 0

        logger.info("Loaded scheduler: %s." % self.name)


    # -------------------------------------------------------------------------
    #
    def pilot_added (self, pilot) :

        pid = pilot.uid

        self.pilots[pid] = dict()
        self.pilots[pid]['resource'] = pilot.resource
        self.pilots[pid]['sandbox']  = pilot.sandbox


    # -------------------------------------------------------------------------
    #
    def pilot_removed (self, pid) :

        if  not pid in self.pilots :
            raise RuntimeError ('cannot remove unknown pilot (%s)' % pid)

        del self.pilots[pid]


    # -------------------------------------------------------------------------
    #
    def schedule(self, units):

        # the scheduler will return a dictionary of the form:
        #   { 
        #     unit_1: pilot_id_1
        #     unit_2: pilot_id_2
        #     unit_4: pilot_id_2
        #     ...
        #   }
        # The scheduler may not be able to schedule some units -- those will
        # simply not be listed for any pilot.  The UM needs to make sure
        # that no UD from the original list is left untreated, eventually.

        pilot_ids          = self.pilots.keys ()
        schedule           = dict()
        schedule['units']  = dict()
        schedule['pilots'] = self.pilots

        if  not len (pilot_ids) :
            raise RuntimeError ('Unit scheduler cannot operate on empty pilot set')


        for unit in units :
            
            if  self.idx >= len(pilot_ids) : 
                self.idx = 0
            
            schedule['units'][unit] = pilot_ids[self.idx]
            self.idx               += 1


        return schedule


    # -------------------------------------------------------------------------




"""
.. module:: radical.pilot.controller.input_file_transfer_worker
.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import time
import saga
import datetime
import traceback
import threading

from bson.objectid import ObjectId
from radical.pilot.states import * 
from radical.pilot.utils.logger import logger

# BULK_LIMIT defines the max. number of transfer requests to pull from DB.
BULK_LIMIT=1

# ----------------------------------------------------------------------------
#
class InputFileTransferWorker(threading.Thread):
    """InputFileTransferWorker handles the staging of input files
    for a UnitManagerController.
    """

    # ------------------------------------------------------------------------
    #
    def __init__(self, session, db_connection_info, unit_manager_id, number=None):

        self._session = session

        # threading stuff
        threading.Thread.__init__(self)
        self.daemon = True

        self.db_connection_info = db_connection_info
        self.unit_manager_id = unit_manager_id

        self._worker_number = number
        self.name = "InputFileTransferWorker-%s" % str(self._worker_number)

        # we cache saga directories for performance, to speed up sandbox
        # creation.
        self._saga_dirs = dict()

    # ------------------------------------------------------------------------
    #
    def run(self):
        """Starts the process when Process.start() is called.
        """

        # make sure to catch sys.exit (which raises SystemExit)
        try :

            logger.info("Starting InputFileTransferWorker")

            # Try to connect to the database and create a tailable cursor.
            try:
                connection = self.db_connection_info.get_db_handle()
                db = connection[self.db_connection_info.dbname]
                um_col = db["%s.cu" % self.db_connection_info.session_id]
                logger.debug("Connected to MongoDB. Serving requests for UnitManager %s." % self.unit_manager_id)

            except Exception, ex:
                logger.error("Connection error: %s. %s" % (str(ex), traceback.format_exc()))
                raise ex

            try :
                while True:
                    # See if we can find a ComputeUnit that is waiting for
                    # input file transfer.
                    compute_unit = None

                    ts = datetime.datetime.utcnow()
                    compute_unit = um_col.find_and_modify(
                        query={"unitmanager": self.unit_manager_id,
                               "FTW_Input_Status": PENDING},
                        update={"$set" : {"FTW_Input_Status": EXECUTING,
                                          "state": STAGING_INPUT},
                                "$push": {"statehistory": {"state": STAGING_INPUT, "timestamp": ts}}},
                        limit=BULK_LIMIT # TODO: bulklimit is probably not the best way to ensure there is just one
                    )
                    state = STAGING_INPUT

                    if compute_unit is None:
                        # Sleep a bit if no new units are available.
                        time.sleep(0.1) # TODO: Probably need better sleep logic as we also have the logic on the end now
                    else:
                        # AM: The code below seems wrong when BULK_LIMIT != 1 -- the
                        # compute_unit will be a list then I assume.
                        try:
                            log_messages = []

                            # We have found a new CU. Now we can process the transfer
                            # directive(s) wit SAGA.
                            compute_unit_id = str(compute_unit["_id"])
                            remote_sandbox = compute_unit["sandbox"]
                            input_staging = compute_unit["FTW_Input_Directives"]

                            # We need to create the cU's directory in case it doesn't exist yet.
                            log_msg = "Creating ComputeUnit sandbox directory %s." % remote_sandbox
                            log_messages.append(log_msg)
                            logger.info(log_msg)

                            # Creating the sandbox directory.
                            try:
                                logger.debug ("saga.fs.Directory ('%s')" % remote_sandbox)

                                remote_sandbox_keyurl = saga.Url (remote_sandbox)
                                remote_sandbox_keyurl.path = '/'
                                remote_sandbox_key = str(remote_sandbox_keyurl)

                                if  remote_sandbox_key not in self._saga_dirs :
                                    self._saga_dirs[remote_sandbox_key] = \
                                            saga.filesystem.Directory (remote_sandbox_key,
                                                    flags=saga.filesystem.CREATE_PARENTS,
                                                    session=self._session)

                                saga_dir = self._saga_dirs[remote_sandbox_key]
                                saga_dir.make_dir (remote_sandbox, 
                                                   flags=saga.filesystem.CREATE_PARENTS)
                            except Exception, ex:
                                tb = traceback.format_exc()
                                logger.info('Error: %s. %s' % (str(ex), tb))

                            logger.info("Processing input file transfers for ComputeUnit %s" % compute_unit_id)
                            # Loop over all transfer directives and execute them.
                            for sd in input_staging:

                                state_doc = um_col.find_one(
                                    {"_id": ObjectId(compute_unit_id)},
                                    fields=["state"]
                                )
                                if state_doc['state'] == CANCELED:
                                    logger.info("Compute Unit Canceled, interrupting input file transfers.")
                                    state = CANCELED
                                    break

                                abs_src = os.path.abspath(sd['source'])
                                input_file_url = saga.Url("file://localhost/%s" % abs_src)
                                if not sd['target']:
                                    target = remote_sandbox
                                else:
                                    target = "%s/%s" % (remote_sandbox, sd['target'])

                                log_msg = "Transferring input file %s -> %s" % (input_file_url, target)
                                logmessage = "Transferred input file %s -> %s" % (input_file_url, target)
                                log_messages.append(log_msg)
                                logger.debug(log_msg)

                                # Execute the transfer.
                                logger.debug ("saga.fs.File ('%s')" % input_file_url)
                                input_file = saga.filesystem.File(
                                    input_file_url,
                                    session=self._session
                                )
                                #try:
                                input_file.copy(target)
                                #except Exception, ex:
                                #    tb = traceback.format_exc()
                                #    logger.info('Error: %s. %s' % (str(ex), tb))

                                input_file.close()

                                # If all went fine, update the state of this StagingDirective to Done
                                um_col.find_and_modify(
                                    query={"_id" : ObjectId(compute_unit_id),
                                           'FTW_Input_Status': EXECUTING,
                                           'FTW_Input_Directives.state': PENDING,
                                           'FTW_Input_Directives.source': sd['source'],
                                           'FTW_Input_Directives.target': sd['target'],
                                           },
                                    update={'$set': {'FTW_Input_Directives.$.state': 'Done'},
                                            '$push': {'log': logmessage}
                                    }
                                )

                        except Exception, ex:
                            # Update the CU's state 'FAILED'.
                            ts = datetime.datetime.utcnow()
                            log_messages = "Input transfer failed: %s\n%s" % (str(ex), traceback.format_exc())
                            logger.error(log_messages)
                            um_col.update(
                                {"_id": ObjectId(compute_unit_id)},
                                {"$set": {"state": FAILED},
                                 "$push": {"statehistory": {"state": FAILED, "timestamp": ts}},
                                 "$push": {"log": log_messages}}
                            )

                    # Code below is only to be run by the "first" or only worker
                    if self._worker_number > 1:
                        continue

                    # If the CU was canceled we can skip the remainder of this loop.
                    if state == CANCELED:
                        continue

                    #
                    # Check to see if there are more pending Directives, if not, we are Done
                    #
                    cursor_w = um_col.find({"unitmanager": self.unit_manager_id,
                                            "$or": [ {"Agent_Input_Status": EXECUTING},
                                                     {"FTW_Input_Status": EXECUTING}
                                                   ]
                                            }
                                           )
                    # Iterate over all the returned CUs (if any)
                    for cu in cursor_w:
                        # See if there are any FTW Input Directives still pending
                        if cu['FTW_Input_Status'] == EXECUTING and \
                                not any(d['state'] == EXECUTING or d['state'] == PENDING for d in cu['FTW_Input_Directives']):
                            # All Input Directives for this FTW are done, mark the CU accordingly
                            um_col.update({"_id": ObjectId(cu["_id"])},
                                          {'$set': {'FTW_Input_Status': DONE},
                                           '$push': {'log': 'All FTW Input Staging Directives done - %d.' % self._worker_number}})

                        # See if there are any Agent Input Directives still pending or executing,
                        # if not, mark it DONE.
                        if cu['Agent_Input_Status'] == EXECUTING and \
                                not any(d['state'] == EXECUTING or d['state'] == PENDING for d in cu['Agent_Input_Directives']):
                            # All Input Directives for this Agent are done, mark the CU accordingly
                            um_col.update({"_id": ObjectId(cu["_id"])},
                                           {'$set': {'Agent_Input_Status': DONE},
                                            '$push': {'log': 'All Agent Input Staging Directives done - %d.' % self._worker_number}
                                           })

                    #
                    # Check for all CUs if both Agent and FTW staging is done, we can then mark the CU PendingExecution
                    #
                    ts = datetime.datetime.utcnow()
                    um_col.find_and_modify(
                        query={"unitmanager": self.unit_manager_id,
                               "Agent_Input_Status": { "$in": [ None, DONE ] },
                               "FTW_Input_Status": { "$in": [ None, DONE ] },
                               "state": STAGING_INPUT
                        },
                        update={"$set": {
                                    "state": PENDING_EXECUTION
                                },
                                "$push": {
                                    "statehistory": {"state": PENDING_EXECUTION, "timestamp": ts}
                                }
                        }
                    )

            except Exception as e :

                logger.error("transfer worker error: %s\n %s" % (str(e), traceback.format_exc()))
                self._session.close (delete=False)
                raise e

        except SystemExit as e :
            logger.error("input file transfer thread caught system exit -- forcing application shutdown")
            import thread
            thread.interrupt_main ()
            





"""
.. module:: radical.pilot.controller.pilot_launcher_worker
.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import time
import saga
import datetime
import traceback
import threading
import radical.utils as ru

from bson.objectid import ObjectId

from radical.pilot.states import * 

from radical.pilot.utils.version import version as VERSION
from radical.pilot.utils.logger import logger
from radical.pilot.context import Context

# BULK_LIMIT defines the max. number of transfer requests to pull from DB.
BULK_LIMIT=1

# The interval at which we check the saga jobs.
JOB_CHECK_INTERVAL   = 60 # seconds
JOB_CHECK_MAX_MISSES =  3 # how often to miss a job be declaring it dead

# ----------------------------------------------------------------------------
#
class PilotLauncherWorker(threading.Thread):
    """PilotLauncherWorker handles bootstrapping and launching of
       the pilot agents.
    """

    # ------------------------------------------------------------------------
    #
    def __init__(self, session, db_connection_info, pilot_manager_id, number=None):
        """Creates a new pilot launcher background process.
        """
        self._session = session

        # threading stuff
        threading.Thread.__init__(self)
        self.daemon = True

        self.db_connection_info = db_connection_info
        self.pilot_manager_id   = pilot_manager_id
        self.name               = "PilotLauncherWorker-%s" % str(number)
        self.missing_pilots     = dict()
        self.job_services       = dict()

    # --------------------------------------------------------------------------
    #
    def check_pilot_states (self, pilot_col) :

        pending_pilots = pilot_col.find(
            {"pilotmanager": self.pilot_manager_id,
             "state"       : {"$in": [PENDING_ACTIVE, ACTIVE]}}
        )

        for pending_pilot in pending_pilots:

            pilot_failed = False
            pilot_done   = False
            reconnected  = False
            pilot_id     = pending_pilot["_id"]
            log_message  = ""
            saga_job_id  = pending_pilot["saga_job_id"]

            logger.info("Performing periodical health check for %s (SAGA job id %s)" % (str(pilot_id), saga_job_id))
            
            if  not pilot_id in self.missing_pilots :
                self.missing_pilots[pilot_id] = 0

            # Create a job service object:
            try: 
                js_url       = saga_job_id.split("]-[")[0][1:]

                if  js_url in self.job_services :
                    js = self.job_services[js_url]
                else :
                    js = saga.job.Service(js_url, session=self._session)
                    self.job_services[js_url] = js

                saga_job     = js.get_job(saga_job_id)
                reconnected  = True

                if  saga_job.state in [saga.job.FAILED, saga.job.CANCELED] :
                    pilot_failed = True
                    log_message  = "SAGA job state for ComputePilot %s is %s."\
                                 % (pilot_id, saga_job.state)

                if  saga_job.state in [saga.job.DONE] :
                    pilot_done = True
                    log_message  = "SAGA job state for ComputePilot %s is %s."\
                                 % (pilot_id, saga_job.state)

            except Exception as e:

                if  not reconnected :
                    logger.warning ('could not reconnect to pilot for state check (%s)' % e)
                    self.missing_pilots[pilot_id] += 1

                    if  self.missing_pilots[pilot_id] >= JOB_CHECK_MAX_MISSES :
                        logger.error ('giving up after 10 attempts')
                        pilot_failed = True
                        log_message  = "Could not reconnect to pilot %s "\
                                       "multiple times - giving up" % pilot_id
                else :
                    logger.warning ('pilot state check failed: %s' % e)
                    pilot_failed = True
                    log_message  = "Couldn't determine job state for ComputePilot %s. " \
                                   "Assuming it has failed." % pilot_id

            if  pilot_failed :
                ts = datetime.datetime.utcnow()
                pilot_col.update(
                    {"_id": pilot_id},
                    {"$set": {"state": FAILED},
                     "$push": {"statehistory": {"state": FAILED, "timestamp": ts}},
                     "$push": {"log": log_message}}
                )
                logger.error (log_message)
                logger.error ('pilot %s declared dead' % pilot_id)

            elif pilot_done :
                # FIXME: this should only be done if the state is not yet
                # done...
                ts = datetime.datetime.utcnow()
                pilot_col.update(
                    {"_id"  : pilot_id,
                     "state": {"$ne"  : DONE}},
                    {"$set" : {"state": DONE},
                     "$push": {"statehistory": {"state": DONE, "timestamp": ts}},
                     "$push": {"log": log_message}}
                )
                logger.error (log_message)
                logger.error ('pilot %s declared dead' % pilot_id)

            else :
                if self.missing_pilots[pilot_id] :
                    logger.info ('pilot %s *assumed* alive and well (%s)' \
                              % (pilot_id, self.missing_pilots[pilot_id]))
                else :
                    logger.info ('pilot %s seems alive and well' \
                              % (pilot_id))


    # ------------------------------------------------------------------------
    #
    def run(self):
        """Starts the process when Process.start() is called.
        """

        # make sure to catch sys.exit (which raises SystemExit)
        try :

            # Try to connect to the database 
            try:
                connection = self.db_connection_info.get_db_handle()
                db = connection[self.db_connection_info.dbname]
                pilot_col = db["%s.p" % self.db_connection_info.session_id]
                logger.debug("Connected to MongoDB. Serving requests for PilotManager %s." % self.pilot_manager_id)

            except Exception, ex:
                tb = traceback.format_exc()
                logger.error("Connection error: %s. %s" % (str(ex), tb))
                return

            last_job_check = time.time()

            while True:
                # Periodically, we pull up all ComputePilots that are pending 
                # execution or were last seen executing and check if the corresponding  
                # SAGA job is still pending in the queue. If that is not the case, 
                # we assume that the job has failed for some reasons and update
                # the state of the ComputePilot accordingly.
                if  last_job_check + JOB_CHECK_INTERVAL < time.time() :
                    last_job_check = time.time()
                    self.check_pilot_states (pilot_col)


                # See if we can find a ComputePilot that is waiting to be launched.
                # If we find one, we use SAGA to create a job service, a job 
                # description and a job that is then send to the local or remote
                # queueing system. If this succedes, we set the ComputePilot's 
                # state to pending, otherwise to failed.
                compute_pilot = None

                ts = datetime.datetime.utcnow()
                compute_pilot = pilot_col.find_and_modify(
                    query={"pilotmanager": self.pilot_manager_id,
                           "state" : PENDING_LAUNCH},
                    update={"$set" : {"state": LAUNCHING},
                            "$push": {"statehistory": {"state": LAUNCHING, "timestamp": ts}}},
                    limit=BULK_LIMIT
                )

                if compute_pilot is None:
                    # Sleep a bit if no new pilots are available.
                    time.sleep(0.1)
                else:
                    try:
                        ######################################################################
                        ##
                        ## LAUNCH THE PILOT AGENT VIA SAGA
                        log_messages = []

                        compute_pilot_id = str(compute_pilot["_id"])
                        logger.info("Launching ComputePilot %s" % compute_pilot)

                        number_cores = compute_pilot['description']['cores']
                        runtime      = compute_pilot['description']['runtime']
                        queue        = compute_pilot['description']['queue']
                        project      = compute_pilot['description']['project']
                        cleanup      = compute_pilot['description']['cleanup']
                        resource_key = compute_pilot['description']['resource']
                        schema       = compute_pilot['description']['access_schema']

                        sandbox      = compute_pilot['sandbox']

                        # check if the user specified a sandbox:
                        if  'sandbox' in compute_pilot['description'] and \
                            compute_pilot['description']['sandbox']   and \
                            compute_pilot['description']['sandbox'] == sandbox :
                            user_sandbox = True
                        else :
                            user_sandbox = False


                        resource_cfg = self._session.get_resource_config(resource_key)
                        agent_worker = resource_cfg.get ('pilot_agent_worker', None)

                        # we expand and exchange keys in the resource config,
                        # depending on the selected schema so better use a deep
                        # copy..
                        import copy
                        resource_cfg = copy.deepcopy (resource_cfg)

                        if  not schema :
                            if 'schemas' in resource_cfg :
                                schema = resource_cfg['schemas'][0]

                        if  not schema in resource_cfg :
                            logger.warning ("schema %s unknown for resource %s -- continue with defaults" \
                                         % (schema, resource_key))

                        else :
                            for key in resource_cfg[schema] :
                                # merge schema specific resource keys into the
                                # resource config
                                resource_cfg[key] = resource_cfg[schema][key]

                        ########################################################
                        # Database connection parameters
                        session_uid  = self.db_connection_info.session_id
                        database_url = self.db_connection_info.dburl

                        surl = saga.Url (database_url)

                        # Set default port if not specified
                        # (explicit is better than implicit!)
                        if not surl.port:
                            surl.port = 27017

                        # Set default host to localhost if not specified
                        if not surl.host:
                            surl.host = 'localhost'

                        database_name = self.db_connection_info.dbname
                        # Set default database name if not specified
                        if not database_name:
                            database_name = 'radicalpilot'

                        database_auth     = self.db_connection_info.dbauth
                        database_hostport = "%s:%d" % (surl.host, surl.port)


                        ########################################################
                        # Get directory where pilot_launcher_worker.py lives
                        plw_dir = os.path.dirname(os.path.realpath(__file__))

                        ########################################################
                        # take 'pilot_agent' as defined in the resource configuration
                        # by default, but override it if set in the Pilot description. 
                        pilot_agent = resource_cfg['pilot_agent']
                        if compute_pilot['description']['pilot_agent_priv'] is not None:
                            pilot_agent = compute_pilot['description']['pilot_agent_priv']
                        agent_path = os.path.abspath("%s/../agent/%s" % (plw_dir, pilot_agent))

                        log_msg = "Using pilot agent %s" % agent_path
                        log_messages.append(log_msg)
                        logger.info(log_msg)

                        ########################################################
                        # we use always "default_bootstrapper.sh" unless a resource 
                        # configuration explicitly defines another bootstrapper. 
                        if 'bootstrapper' in resource_cfg and resource_cfg['bootstrapper'] is not None:
                            bootstrapper = resource_cfg['bootstrapper']
                        else:
                            bootstrapper = 'default_bootstrapper.sh'
                        bootstrapper_path = os.path.abspath("%s/../bootstrapper/%s" % (plw_dir, bootstrapper))
                        
                        log_msg = "Using bootstrapper %s" % bootstrapper_path
                        log_messages.append(log_msg)
                        logger.info(log_msg)

                        ########################################################
                        # Create SAGA Job description and submit the pilot job #
                        ########################################################

                        # log_msg = "Creating agent sandbox '%s'." % str(sandbox)
                        # log_messages.append(log_msg)
                        # logger.debug(log_msg)
                        #
                        # logger.debug ("saga.fs.Directory ('%s')" % saga.Url(sandbox))
                        # agent_dir = saga.filesystem.Directory(
                        #     saga.Url(sandbox),
                        #     saga.filesystem.CREATE_PARENTS, session=self._session)
                        # agent_dir.close()

                        ########################################################
                        # Copy the bootstrap shell script.  This also creates
                        # the sandbox
                        bs_script_url = saga.Url("file://localhost/%s" % bootstrapper_path)
                        bs_script_tgt = saga.Url("%s/%s"               % (sandbox, bootstrapper))
                        log_msg = "Copying bootstrapper '%s' to agent sandbox (%s)." % (bs_script_url, bs_script_tgt)
                        log_messages.append(log_msg)
                        logger.debug(log_msg)

                        bs_script = saga.filesystem.File(bs_script_url, session=self._session)
                        bs_script.copy(bs_script_tgt, flags=saga.filesystem.CREATE_PARENTS)
                        bs_script.close()

                        ########################################################
                        # Copy the agent script
                        agent_script_url = saga.Url("file://localhost/%s" % agent_path)
                        log_msg = "Copying agent '%s' to agent sandbox (%s)." % (agent_script_url, sandbox)
                        log_messages.append(log_msg)
                        logger.debug(log_msg)

                        agent_script = saga.filesystem.File(agent_script_url)
                        agent_script.copy("%s/radical-pilot-agent.py" % str(sandbox))
                        agent_script.close()

                        # copying agent-worker.py script to sandbox
                        #########################################################

                        if agent_worker is not None:
                            logger.warning("Using custom agent worker script: %s" % agent_worker)
                            worker_path = os.path.abspath("%s/../agent/%s" % (plw_dir, agent_worker))

                            worker_script_url = saga.Url("file://localhost/%s" % worker_path)

                            log_msg = "Copying '%s' to agent sandbox (%s)." % (worker_script_url, sandbox)
                            log_messages.append(log_msg)
                            logger.debug(log_msg)

                            worker_script = saga.filesystem.File(worker_script_url)
                            worker_script.copy("%s/agent-worker.py" % str(sandbox))
                            worker_script.close()

                        #########################################################
                        # now that the script is in place and we know where it is,
                        # we can launch the agent
                        job_service_url = saga.Url(resource_cfg['job_manager_endpoint'])
                        logger.debug ("saga.job.Service ('%s')" % job_service_url)
                        if  job_service_url in self.job_services :
                            js = self.job_services[job_service_url]
                        else :
                            js = saga.job.Service(job_service_url, session=self._session)
                            self.job_services[job_service_url] = js

                        jd = saga.job.Description()
                        jd.working_directory = saga.Url(sandbox).path

                        bootstrap_args = "-n %s -s %s -p %s -t %s -d %s -c %s -v %s" %\
                            (database_name, session_uid, str(compute_pilot_id),
                             runtime, logger.level, number_cores, VERSION)

                        if  user_sandbox :
                            bootstrap_args += " -u"

                        if 'agent_mongodb_endpoint' in resource_cfg and resource_cfg['agent_mongodb_endpoint'] is not None:
                            agent_db_url = ru.Url(resource_cfg['agent_mongodb_endpoint'])
                            bootstrap_args += " -m %s:%d " % (agent_db_url.host, agent_db_url.port)
                        else:
                            bootstrap_args += " -m %s " % database_hostport
 
                        bootstrap_args += " -a %s " % database_auth


                        if 'python_interpreter' in resource_cfg and resource_cfg['python_interpreter'] is not None:
                            bootstrap_args += " -i %s " % resource_cfg['python_interpreter']
                        if 'pre_bootstrap' in resource_cfg and resource_cfg['pre_bootstrap'] is not None:
                            for command in resource_cfg['pre_bootstrap']:
                                bootstrap_args += " -e '%s' " % command
                        if 'global_virtenv' in resource_cfg and resource_cfg['global_virtenv'] is not None:
                            bootstrap_args += " -g %s " % resource_cfg['global_virtenv']
                        if 'lrms' in resource_cfg and resource_cfg['lrms'] is not None:
                            bootstrap_args += " -l %s " % resource_cfg['lrms']
                        else:
                            raise Exception("LRMS not specified.")
                        if 'task_launch_method' in resource_cfg and resource_cfg['task_launch_method'] is not None:
                            bootstrap_args += " -j %s " % resource_cfg['task_launch_method']
                        else:
                            raise Exception("Task launch method not set.")
                        if 'mpi_launch_method' in resource_cfg and resource_cfg['mpi_launch_method'] is not None:
                            bootstrap_args += " -k %s " % resource_cfg['mpi_launch_method']
                        else:
                            raise Exception("MPI launch method not set.")
                        if 'forward_tunnel_endpoint' in resource_cfg and resource_cfg['forward_tunnel_endpoint'] is not None:
                            bootstrap_args += " -f %s " % resource_cfg['forward_tunnel_endpoint']

                        if cleanup is True: 
                            # cleanup flags:
                            #   l : pilot log files
                            #   u : unit work dirs
                            #   v : virtualenv
                            #   e : everything (== pilot sandbox)
                            # FIXME: get cleanup flags from somewhere
                            logger.info ('request cleanup for pilot %s' % compute_pilot_id)
                            bootstrap_args += " -x %s" % 'luve' # the cleanup flag

                        if  'RADICAL_PILOT_BENCHMARK' in os.environ :
                            bootstrap_args += " -b"

                        jd.executable = "/bin/bash"
                        jd.arguments = ["-l", bootstrapper, bootstrap_args]

                        logger.debug("Bootstrap command line: %s %s" % (jd.executable, jd.arguments))

                        # fork:// and ssh:// don't support 'queue' and 'project'
                        if (job_service_url.schema != "fork") and (job_service_url.schema != "ssh"):

                            # process the 'queue' attribute
                            if queue is not None:
                                jd.queue = queue
                            elif 'default_queue' in resource_cfg:
                                jd.queue = resource_cfg['default_queue']

                            # process the project / allocation 
                            if project is not None:
                                jd.project = project

                        # set the SPMD variation if required
                        if 'spmd_variation' in resource_cfg and resource_cfg['spmd_variation'] is not None:
                            jd.spmd_variation = resource_cfg['spmd_variation']

                        jd.output = "AGENT.STDOUT"
                        jd.error  = "AGENT.STDERR"
                        jd.total_cpu_count = number_cores
                        jd.wall_time_limit = runtime
                        if compute_pilot['description']['memory'] is not None:
                            jd.total_physical_memory = compute_pilot['description']['memory']

                        log_msg = "Submitting SAGA job with description: %s" % str(jd.as_dict())
                        log_messages.append(log_msg)
                        logger.debug(log_msg)

                        pilotjob = js.create_job(jd)
                        pilotjob.run()

                        # do a quick error check
                        if pilotjob.state == saga.FAILED:
                            raise Exception("SAGA Job state was FAILED.")

                        saga_job_id = pilotjob.id
                        log_msg = "SAGA job submitted with job id %s" % str(saga_job_id)

                        log_messages.append(log_msg)
                        logger.debug(log_msg)

                        ##
                        ##
                        ######################################################################

                        # Update the Pilot's state to 'PENDING_ACTIVE' if SAGA job submission was successful.
                        ts = datetime.datetime.utcnow()
                        ret = pilot_col.update(
                            {"_id"  : ObjectId(compute_pilot_id),
                             "state": 'Launching'},
                            {"$set": {"state": PENDING_ACTIVE,
                                      "saga_job_id": saga_job_id},
                             "$push": {"statehistory": {"state": PENDING_ACTIVE, "timestamp": ts}},
                             "$pushAll": {"log": log_messages}}                    
                        )

                        if  ret['n'] == 0 :
                            # could not update, probably because the agent is
                            # running already.  Just update state history and
                            # jobid then
                            ret = pilot_col.update(
                                {"_id"  : ObjectId(compute_pilot_id)},
                                {"$set" : {"saga_job_id": saga_job_id},
                                 "$push": {"statehistory": {"state": PENDING_ACTIVE, "timestamp": ts}},
                                 "$pushAll": {"log": log_messages}}                    
                            )


                    except Exception, ex:
                        # Update the Pilot's state 'FAILED'.
                        ts = datetime.datetime.utcnow()
                        log_messages = "Pilot launching failed: %s\n%s" % (str(ex), traceback.format_exc())
                        pilot_col.update(
                            {"_id": ObjectId(compute_pilot_id)},
                            {"$set": {"state": FAILED},
                             "$push": {"statehistory": {"state": FAILED, "timestamp": ts}},
                             "$push": {"log": log_messages}}
                        )
                        logger.error(log_messages)

        except SystemExit as e :
            logger.exception("pilot launcher thread caught system exit -- forcing application shutdown")
            import thread
            thread.interrupt_main ()
            




"""
.. module:: radical.pilot.controller.pilot_launcher_worker
.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import time
import Queue
import weakref
import datetime
import threading

from multiprocessing import Pool

from radical.utils        import which
from radical.utils        import Url
from radical.pilot.types  import *
from radical.pilot.states import *
from radical.pilot.utils.logger import logger

from radical.pilot.controller.input_file_transfer_worker import InputFileTransferWorker
from radical.pilot.controller.output_file_transfer_worker import OutputFileTransferWorker

from radical.pilot.staging_directives import TRANSFER, LINK, COPY, MOVE

# ----------------------------------------------------------------------------
#
class UnitManagerController(threading.Thread):
    """UnitManagerController handles backend interaction for the UnitManager 
    class. It is threaded and manages background worker processes. 
    """

    # ------------------------------------------------------------------------
    #
    def __init__(self, unit_manager_uid, session, db_connection, db_connection_info,
        scheduler=None, input_transfer_workers=None,
        output_transfer_workers=None):

        self._session = session

        # Multithreading stuff
        threading.Thread.__init__(self)
        self.daemon = True

        # Stop event can be set to terminate the main loop
        self._stop = threading.Event()
        self._stop.clear()

        # Initialized is set, once the run loop has pulled status
        # at least once. Other functions use it as a guard.
        self._initialized = threading.Event()
        self._initialized.clear()

        # The shard_data_manager handles data exchange between the worker
        # process and the API objects. The communication is unidirectional:
        # workers WRITE to _shared_data and API methods READ from _shared_data.
        # The strucuture of _shared_data is as follows:
        #
        # { unit1_uid: MongoDB document (dict),
        #   unit2_uid: MongoDB document (dict),
        #   ...
        # }
        #
        self._shared_data = dict()
        self._shared_data_lock = threading.Lock()

        # The manager-level list.
        #
        self._manager_callbacks = dict()

        # The MongoDB database handle.
        self._db = db_connection

        if unit_manager_uid is None:
            # Try to register the UnitManager with the database.
            self._um_id = self._db.insert_unit_manager(
                scheduler=scheduler,
                input_transfer_workers=input_transfer_workers,
                output_transfer_workers=output_transfer_workers)
            self._num_input_transfer_workers = input_transfer_workers
            self._num_output_transfer_workers = output_transfer_workers
        else:
            um_json = self._db.get_unit_manager(unit_manager_id=unit_manager_uid)
            self._um_id = unit_manager_uid
            self._num_input_transfer_workers = um_json["input_transfer_workers"]
            self._num_output_transfer_workers = um_json["output_transfer_workers"]

        # The INPUT transfer worker(s) are autonomous processes that
        # execute input file transfer requests concurrently.
        self._input_file_transfer_worker_pool = []
        for worker_number in range(1, self._num_input_transfer_workers+1):
            worker = InputFileTransferWorker(
                session=self._session,
                db_connection_info=db_connection_info, 
                unit_manager_id=self._um_id,
                number=worker_number
            )
            self._input_file_transfer_worker_pool.append(worker)
            worker.start()

        # The OUTPUT transfer worker(s) are autonomous processes that
        # execute output file transfer requests concurrently.
        self._output_file_transfer_worker_pool = []
        for worker_number in range(1, self._num_output_transfer_workers+1):
            worker = OutputFileTransferWorker(
                session=self._session,
                db_connection_info=db_connection_info, 
                unit_manager_id=self._um_id,
                number=worker_number
            )
            self._output_file_transfer_worker_pool.append(worker)
            worker.start()

        self._callback_histories = dict ()

    # ------------------------------------------------------------------------
    #
    @classmethod
    def uid_exists(cls, db_connection, unit_manager_uid):
        """Checks wether a particular unit manager UID exists.
        """
        exists = False

        if unit_manager_uid in db_connection.list_unit_manager_uids():
            exists = True

        return exists

    # ------------------------------------------------------------------------
    #
    @property
    def unit_manager_uid(self):
        """Returns the uid of the associated UnitManager
        """
        return self._um_id

    # ------------------------------------------------------------------------
    #
    def stop(self):
        """stop() signals the process to finish up and terminate.
        """
        self._stop.set()
        self.join()
        logger.debug("Worker thread (ID: %s[%s]) for UnitManager %s stopped." %
                    (self.name, self.ident, self._um_id))

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_data(self, unit_uid):
        """Retruns the raw data (json dicts) of one or more ComputeUnits
           registered with this Worker / UnitManager
        """
        # Wait for the initialized event to assert proper operation.
        self._initialized.wait()

        return self._shared_data[unit_uid]["data"]

    # ------------------------------------------------------------------------
    #
    def call_unit_state_callbacks(self, unit_id, new_state):
        """Wrapper function to call all all relevant callbacks, on unit-level
        as well as manager-level.
        """

        # this is the point where, at the earliest, the application could have
        # been notified about unit state changes.  So we record that event.
        if  not unit_id in self._callback_histories :
            self._callback_histories[unit_id] = list()
        self._callback_histories[unit_id].append (
                {'timestamp' : datetime.datetime.utcnow(), 
                 'state'     : new_state})
    
        for cb in self._shared_data[unit_id]['callbacks']:
            try:

                if self._shared_data[unit_id]['facade_object'] :
                    cb(self._shared_data[unit_id]['facade_object'], new_state)
                else :
                    logger.error("Couldn't call callback (no pilot instance)")
            except Exception, ex:
                logger.error(
                    "Couldn't call callback function %s" % str(ex))
                raise

        # If we have any manager-level callbacks registered, we
        # call those as well!
        if  not UNIT_STATE in self._manager_callbacks :
            self._manager_callbacks[UNIT_STATE] = list()

        for cb in self._manager_callbacks[UNIT_STATE]:
            try:
                cb(self._shared_data[unit_id]['facade_object'],
                   new_state)
            except Exception, ex:
                logger.error(
                    "Couldn't call callback function %s" % str(ex))
                raise

        # If we meet a final state, we record the object's callback history for
        # later evaluation.
        if  new_state in (DONE, FAILED, CANCELED) :
            self._db.publish_compute_unit_callback_history (unit_id, self._callback_histories[unit_id])


    # ------------------------------------------------------------------------
    #
    # def _set_state(self, unit_uid, state, log):

    #     if not isinstance(log, list):
    #         log = [log]

    #     # Acquire the shared data lock.
    #     self._shared_data_lock.acquire()

    #     old_state = self._shared_data[unit_uid]["data"]["state"]

    #     # Update the database.
    #     self._db.set_compute_unit_state(unit_uid, state, log)

    #     # Update shared data.
    #     self._shared_data[unit_uid]["data"]["state"] = state
    #     self._shared_data[unit_uid]["data"]["statehistory"].append(state)
    #     self._shared_data[unit_uid]["data"]["log"].extend(log)

    #     # Call the callbacks
    #     if state != old_state:
    #         # On a state change, we fire zee callbacks.
    #         logger.info(
    #             "XX ComputeUnit '%s' state changed from '%s' to '%s'." %
    #             (unit_uid, old_state, state)
    #         )

    #         # The state of the unit has changed, We call all
    #         # unit-level callbacks to propagate this.
    #         self.call_unit_state_callbacks(unit_uid, state)

    #     # Release the shared data lock.
    #     self._shared_data_lock.release()

    # ------------------------------------------------------------------------
    #
    def run(self):
        """run() is called when the process is started via
           PilotManagerController.start().
        """

        # make sure to catch sys.exit (which raises SystemExit)
        try :

            logger.debug("Worker thread (ID: %s[%s]) for UnitManager %s started." %
                        (self.name, self.ident, self._um_id))

            # transfer results contains the futures to the results of the
            # asynchronous transfer operations.
            transfer_results = list()

            while not self._stop.is_set():

                # =================================================================
                #
                # Check and update units. This needs to be optimized at
                # some point, i.e., state pulling should be conditional
                # or triggered by a tailable MongoDB cursor, etc.
                unit_list = self._db.get_compute_units(unit_manager_id=self._um_id)


                for unit in unit_list:
                    unit_id = str(unit["_id"])

                    new_state = unit["state"]
                    if unit_id in self._shared_data:
                        old_state = self._shared_data[unit_id]["data"]["state"]
                    else:
                        old_state = None
                        self._shared_data_lock.acquire()
                        self._shared_data[unit_id] = {
                            'data':          unit,
                            'callbacks':     [],
                            'facade_object': None
                        }
                        self._shared_data_lock.release()

                    self._shared_data_lock.acquire()
                    self._shared_data[unit_id]["data"] = unit
                    self._shared_data_lock.release()

                    if new_state != old_state:
                        # On a state change, we fire zee callbacks.
                        logger.info("RUN ComputeUnit '%s' state changed from '%s' to '%s'." % (unit_id, old_state, new_state))

                        # The state of the unit has changed, We call all
                        # unit-level callbacks to propagate this.
                        self.call_unit_state_callbacks(unit_id, new_state)

                # After the first iteration, we are officially initialized!
                if not self._initialized.is_set():
                    self._initialized.set()

                # sleep a little if this cycle was idle
                if  not len(unit_list) :
                    time.sleep(0.1)


            # shut down the autonomous input / output transfer worker(s)
            for worker in self._input_file_transfer_worker_pool:
              # worker.terminate()
                logger.debug("UnitManager.close(): %s terminated." % worker.name)
              # worker.join()

            for worker in self._output_file_transfer_worker_pool:
              # worker.terminate()
                logger.debug("UnitManager.close(): %s terminated." % worker.name)
              # worker.join()
        except SystemExit as e :
            logger.exception ("unit manager controller thread caught system exit -- forcing application shutdown")
            import thread
            thread.interrupt_main ()
            

    # ------------------------------------------------------------------------
    #
    def register_unit_callback(self, unit, callback_func):
        """Registers a callback function for a ComputeUnit.
        """
        unit_uid = unit.uid

        self._shared_data_lock.acquire()
        self._shared_data[unit_uid]['callbacks'].append(callback_func)
        self._shared_data_lock.release()

        # Add the facade object if missing, e.g., after a re-connect.
        if self._shared_data[unit_uid]['facade_object'] is None:
            self._shared_data_lock.acquire()
            self._shared_data[unit_uid]['facade_object'] = unit # weakref.ref(unit)
            self._shared_data_lock.release()

        # Callbacks can only be registered when the ComputeUnit lready has a
        # state. To partially address this shortcomming we call the callback
        # with the current ComputeUnit state as soon as it is registered.
        self.call_unit_state_callbacks(
            unit_uid,
            self._shared_data[unit_uid]["data"]["state"]
        )

    # ------------------------------------------------------------------------
    #
    def register_manager_callback(self, callback_func, metric):
        """Registers a manager-level callback.
        """
        if not metric in self._manager_callbacks :
            self._manager_callbacks[metric] = list()

        self._manager_callbacks[metric].append(callback_func)

    # ------------------------------------------------------------------------
    #
    def fire_manager_callback(self, metric, obj, value):
        """Fire a manager-level callback.
        """
        if  not metric in self._manager_callbacks :
            self._manager_callbacks[metric] = list()

        for cb in self._manager_callbacks[metric] :
            try:
                cb (obj, value)
            except Exception, ex:
                logger.error ("Couldn't call '%s' callback function %s: %s" \
                           % (metric, cb, ex))
                raise

    # ------------------------------------------------------------------------
    #
    def get_unit_manager_data(self):
        """Returns the raw data (JSON dict) for a UnitManger.
        """
        return self._db.get_unit_manager(self._um_id)

    # ------------------------------------------------------------------------
    #
    def get_pilot_uids(self):
        """Returns the UIDs of the pilots registered with the UnitManager.
        """
        return self._db.unit_manager_list_pilots(self._um_id)

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_uids(self):
        """Returns the UIDs of all ComputeUnits registered with the
        UnitManager.
        """
        return self._db.unit_manager_list_compute_units(self._um_id)

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_states(self, unit_uids=None):
        """Returns the states of all ComputeUnits registered with the
        Unitmanager.
        """
        return self._db.get_compute_unit_states(
            self._um_id, unit_uids)

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_stdout(self, compute_unit_uid):
        """Returns the stdout for a compute unit.
        """
        return self._db.get_compute_unit_stdout(compute_unit_uid)

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_stderr(self, compute_unit_uid):
        """Returns the stderr for a compute unit.
        """
        return self._db.get_compute_unit_stderr(compute_unit_uid)

    # ------------------------------------------------------------------------
    #
    def add_pilots(self, pilots):
        """Links ComputePilots to the UnitManager.
        """
        # Extract the uids
        pids = []
        for pilot in pilots:
            pids.append(pilot.uid)

        self._db.unit_manager_add_pilots(unit_manager_id=self._um_id,
                                         pilot_ids=pids)

    # ------------------------------------------------------------------------
    #
    def remove_pilots(self, pilot_uids):
        """Unlinks one or more ComputePilots from the UnitManager.
        """
        self._db.unit_manager_remove_pilots(unit_manager_id=self._um_id,
                                            pilot_ids=pilot_uids)

    # ------------------------------------------------------------------------
    #
    def publish_compute_units(self, units):
        """register the unscheduled units in the database"""

        # Add all units to the database.
        results = self._db.insert_compute_units(
            unit_manager_uid=self._um_id,
            units=units,
            unit_log=[]
        )

        assert len(units) == len(results)

        # Match results with units.
        for unit in units:
            # Create a shared data store entry
            self._shared_data[unit.uid] = {
                'data':          results[unit.uid],
                'callbacks':     [],
                'facade_object': unit # weakref.ref(unit)
            }


    # ------------------------------------------------------------------------
    #
    def schedule_compute_units(self, pilot_uid, units):
        """Request the scheduling of one or more ComputeUnits on a
           ComputePilot.
        """

        try:
            cu_transfer   = list()
            cu_notransfer = list()

            # Get some information about the pilot sandbox from the database.
            pilot_info = self._db.get_pilots(pilot_ids=pilot_uid)
            # TODO: this hack below relies on what?! That there is just one pilot?
            pilot_sandbox = pilot_info[0]['sandbox']

            # Split units into two different lists: the first list contains the CUs
            # that need file transfer and the second list contains the CUs that
            # don't. The latter is added to the pilot directly, while the former
            # is added to the transfer queue.
            for unit in units:

                # Create object for staging status tracking
                unit.FTW_Input_Status = None
                unit.FTW_Input_Directives = []
                unit.Agent_Input_Status = None
                unit.Agent_Input_Directives = []
                unit.FTW_Output_Status = None
                unit.FTW_Output_Directives = []
                unit.Agent_Output_Status = None
                unit.Agent_Output_Directives = []

                # Split the input staging directives over the transfer worker and the agent
                input_sds = unit.description.input_staging
                if not isinstance(input_sds, list):
                    # Ugly, but is a workaround for iterating on attribute interface
                    # TODO: Verify if this piece of code is actually still required
                    if input_sds:
                        input_sds = [input_sds]
                    else:
                        input_sds = []
                for input_sd_entry in input_sds:

                    action = input_sd_entry['action']
                    source = Url(input_sd_entry['source'])
                    target = Url(input_sd_entry['target'])

                    new_sd = {'action':   action,
                              'source':   str(source),
                              'target':   str(target),
                              'flags':    input_sd_entry['flags'],
                              'priority': input_sd_entry['priority'],
                              'state':    PENDING
                    }

                    if action == LINK or action == COPY or action == MOVE:
                        unit.Agent_Input_Directives.append(new_sd)
                        unit.Agent_Input_Status = PENDING
                    elif action == TRANSFER:
                        if source.scheme and source.scheme != 'file':
                            # If there is a scheme and it is different than "file",
                            # assume a remote pull from the agent
                            unit.Agent_Input_Directives.append(new_sd)
                            unit.Agent_Input_Status = PENDING
                        else:
                            # Transfer from local to sandbox
                            unit.FTW_Input_Directives.append(new_sd)
                            unit.FTW_Input_Status = PENDING
                    else:
                        logger.error('Not sure if action %s makes sense for input staging' % action)

                # Split the output staging directives over the transfer worker and the agent
                output_sds = unit.description.output_staging
                if not isinstance(output_sds, list):
                    # Ugly, but is a workaround for iterating on att iface
                    # TODO: Verify if this piece of code is actually still required
                    if output_sds:
                        output_sds = [output_sds]
                    else:
                        output_sds = []
                for output_sds_entry in output_sds:

                    action = output_sds_entry['action']
                    source = Url(output_sds_entry['source'])
                    target = Url(output_sds_entry['target'])

                    new_sd = {'action':   action,
                              'source':   str(source),
                              'target':   str(target),
                              'flags':    output_sds_entry['flags'],
                              'priority': output_sds_entry['priority'],
                              'state':    PENDING
                    }

                    if action == LINK or action == COPY or action == MOVE:
                        unit.Agent_Output_Directives.append(new_sd)
                        unit.Agent_Output_Status = NEW
                    elif action == TRANSFER:
                        if target.scheme and target.scheme != 'file':
                            # If there is a scheme and it is different than "file",
                            # assume a remote push from the agent
                            unit.Agent_Output_Directives.append(new_sd)
                            unit.Agent_Output_Status = NEW
                        else:
                            # Transfer from sandbox back to local
                            unit.FTW_Output_Directives.append(new_sd)
                            unit.FTW_Output_Status = NEW
                    else:
                        logger.error('Not sure if action %s makes sense for output staging' % action)

                if unit.FTW_Input_Directives or unit.Agent_Input_Directives:
                    log = ["Scheduled for data transfer to ComputePilot %s." % pilot_uid]
                    self._db.set_compute_unit_state(unit.uid, PENDING_INPUT_STAGING, log)
                    cu_transfer.append(unit)
                else:
                    cu_notransfer.append(unit)

            # Bulk-add all non-transfer units-
            self._db.assign_compute_units_to_pilot(
                units=cu_notransfer,
                pilot_uid=pilot_uid,
                pilot_sandbox=pilot_sandbox
            )

            self._db.assign_compute_units_to_pilot(
                units=cu_transfer,
                pilot_uid=pilot_uid,
                pilot_sandbox=pilot_sandbox
            )

            for unit in cu_notransfer:
                log = ["Scheduled for execution on ComputePilot %s." % pilot_uid]
                self._db.set_compute_unit_state(unit.uid, PENDING_EXECUTION, log)
                #self._set_state(uid, PENDING_EXECUTION, log)

            logger.info(
                "Scheduled ComputeUnits %s for execution on ComputePilot '%s'." %
                (cu_notransfer, pilot_uid)
            )
        except Exception, e:
            import traceback
            logger.error (traceback.format_exc())
            raise Exception('error in unit manager controler: %s' % e)




#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.scheduler.BackfillingScheduler
   :platform: Unix
   :synopsis: A multi-pilot, backfilling scheduler.

.. moduleauthor:: Mark Santcroos <mark.santcroos@rutgers.edu>
"""

__copyright__ = "Copyright 2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os 
import pprint

from radical.pilot.utils.logger        import logger
from radical.pilot.scheduler.interface import Scheduler
from radical.pilot.states              import *

# -----------------------------------------------------------------------------
# 
class BackfillingScheduler(Scheduler):
    """
    
    BackfillingScheduler implements a multi-pilot, backfilling scheduling
    algorithm. Only schedules CUs to Pilots that are active and have
    a free-slot.

    This scheduler is not able to handle pilots which serve more than one unit
    manager concurrently.

    """

    # -------------------------------------------------------------------------
    #
    def __init__ (self, manager, session):
        """
        """
        logger.info("Loaded scheduler: %s." % self.name)

        self.manager = manager
        self.session = session
        self.waitq   = list()
        self.runq    = list()
        self.pmgrs   = list()
        self.pilots  = dict()

        # make sure the UM notifies us on all unit state changes
        manager.register_callback (self._unit_state_callback)


    # -------------------------------------------------------------------------
    #
    def _update_caps (self) :

        # we keep track of available cores on our own.  However, we should sync
        # our bookkeeping with reality now and then...  
        #
        # NOTE: Not sure yet when this methid will be called -- calling it too
        # frequently will slow scheduling down significantly, calling it very
        # infrequently can result in invalid schedules.  Accuracy vs.
        # performance...

        pilot_docs = self.manager._worker._db.get_pilots (pilot_ids=self.pilots.keys ())

        for pilot_doc in pilot_docs :

            pid = str (pilot_doc['_id'])
            if  not pid in self.pilots :
                raise RuntimeError ("Got invalid pilot doc (%s)" % pid)

            self.pilots[pid]['state'] = str(pilot_doc.get ('state'))
            self.pilots[pid]['cap']   = int(pilot_doc.get ('capability', 0))


    # -------------------------------------------------------------------------
    #
    def _unit_state_callback (self, unit, state) :
        
        try :

            uid = unit.uid

            if  not unit in self.runq :
                # as we cannot unregister callbacks, we simply ignore this
                # invokation.  Its probably from a unit we handled previously.
                # (although this should have been final?)
                # FIXME: how can I *un*register a unit callback?
                return

            logger.debug ("[SchedulerCallback]: Computeunit %s changed to %s" % (uid, state))

            if  state in [PENDING_OUTPUT_STAGING, STAGING_OUTPUT, DONE, FAILED, CANCELED] :
                # the pilot which owned this CU should now have free slots available
                # FIXME: how do I get the pilot from the CU?
                
                pid = unit.execution_details.get ('pilot', None)

                if  not pid :
                    logger.warning ('cannot handle final unit %s w/o pilot information' % uid)

                if  pid not in self.pilots :
                    logger.warning ('cannot handle unit %s cb for pilot %s (pilot is gone)' % (uid, pid))

                elif unit in self.runq :
                    # only interpret this event once, on any of the states above,
                    # whichever occurs first
                    self.pilots[pid]['caps'] += unit.description.cores
                    self.runq.remove (unit)
                    self._reschedule (pid=pid)
              #     logger.debug ('unit %s frees %s cores on (-> %s)' \
              #                % (uid, unit.description.cores, pid, self.pilots[pid]['caps']))

              # else :
              #     logger.debug ('unit %s freed %s cores on %s (== %s) -- not reused!'
              #                % (uid, unit.description.cores, pid, self.pilots[pid]['caps']))

        except Exception as e :
            logger.error ("error in unit callback for backfiller (%s) - ignored" % e)


    # -------------------------------------------------------------------------
    #
    def _pilot_state_callback (self, pilot, state) :
        
        try :

            pid = pilot.uid
    
            if  not pid in self.pilots :
                # as we cannot unregister callbacks, we simply ignore this
                # invokation.  Its probably from a pilot we used previously.
                logger.warn ("[SchedulerCallback]: ComputePilot %s changed to %s (ignored)" % (pid, state))
                return
    
    
            self.pilots[pid]['state'] = state
            logger.debug ("[SchedulerCallback]: ComputePilot %s changed to %s" % (pid, state))
    
            if  state in [ACTIVE] :
                # the pilot is now ready to be used
                self._reschedule (pid=pid)
    
            if  state in [DONE, FAILED, CANCELED] :
                # we can't use this pilot anymore...  
                del self.pilots[pid]
    
                # FIXME: how can I *un*register a pilot callback?
    
        except Exception as e :
          # import traceback
          # traceback.print_exc ()
            logger.error ("error in pilot callback for backfiller (%s) - ignored" % e)


    # -------------------------------------------------------------------------
    #
    def pilot_added (self, pilot) :

        pid = pilot.uid

        # get initial information about the pilot capabilities
        #
        # NOTE: this assumes that the pilot manages no units, yet.  This will
        # generally be true, as the UM will call this methods before it submits
        # any units.  This will, however, work badly with pilots which are added
        # to more than one UM.  This though holds true for other parts in this
        # code as well, thus we silently ignore this issue for now, and accept
        # this as known limitation....
        self.pilots[pid] = dict()
        self.pilots[pid]['cores']    = pilot.description.cores
        self.pilots[pid]['caps']     = pilot.description.cores
        self.pilots[pid]['state']    = pilot.state
        self.pilots[pid]['resource'] = pilot.resource
        self.pilots[pid]['sandbox']  = pilot.sandbox

        # make sure we register callback only once per pmgr
        pmgr = pilot.pilot_manager
        if  pmgr not in self.pmgrs :
            self.pmgrs.append (pmgr)
            pmgr.register_callback (self._pilot_state_callback)

        # if we have any pending units, we better serve them now...
        self._reschedule (pid=pid)


    # -------------------------------------------------------------------------
    #
    def pilot_removed (self, pid) :

        if  not pid in self.pilots :
            raise RuntimeError ('cannot remove unknown pilot (%s)' % pid)

        del self.pilots[pid]
        # FIXME: how can I *un*register a pilot callback?

        # no need to schedule, really


    # -------------------------------------------------------------------------
    #
    def schedule (self, units) :

        # this call really just adds the incoming units to the wait queue and
        # then calls reschedule() to have them picked up.
        for unit in units :
            
            if  unit in self.waitq :
                raise RuntimeError ('Unit cannot be scheduled twice (%s)' % unit.uid)

            if  unit in self.runq :
                raise RuntimeError ('Unit cannot be scheduled twice (%s)' % unit.uid)

            if  unit.state != NEW :
                raise RuntimeError ('Unit %s not in NEW state (%s)' % unit.uid)

            self.waitq.append (unit)

        # lets see what we can do about the known units...
        self._reschedule ()


    
    # -------------------------------------------------------------------------
    #
    def unschedule (self, units) :

        # the UM revokes the control over this unit from us...

        for unit in units :

            uid = unit.uid

            if  unit in self.runq  :
                raise RuntimeError ('cannot unschedule assigned unit (%s)' % uid)

            if  not unit in self.waitq :
                raise RuntimeError ('cannot remove unknown unit (%s)' % uid)

            # NOTE: we don't care if that pilot had any CUs active -- its up to the
            # UM what happens to those.

            self.waitq.remove (unit)
            # FIXME: how can I *un*register a pilot callback?
            # FIXME: is this is a race condition with the unit state callback
            #        actions on the queues?


    # -------------------------------------------------------------------------
    #
    def _reschedule (self, pid=None) :

        # dig through the list of waiting CUs, and try to find a pilot for each
        # of them.  This enacts first-come-first-served, but will be unbalanced
        # if the units in the queue are of different sizes (it is kind of
        # opposite to backfilling -- its frontfilling ;).  That problem is
        # ignored at this point.
        #
        # if any units get scheduled, we push a dictionary to the UM to enact
        # the schedule:
        #   { 
        #     unit_1: [pilot_id_1, pilot_resource_name]
        #     unit_2: [pilot_id_2, pilot_resource_name]
        #     unit_4: [pilot_id_2, pilot_resource_name]
        #     ...
        #   }

        if  not len(self.pilots.keys ()) :
            # no pilots to  work on, yet.
            logger.warning ("cannot schedule -- no pilots available")
            return 

        if  pid and pid not in self.pilots :
            logger.warning ("cannot schedule -- invalid target pilot %s" % pid)
            raise RuntimeError ("Invalid pilot (%s)" % pid)
            

        schedule           = dict()
        schedule['units']  = dict()
        schedule['pilots'] = self.pilots

        logger.debug ("schedule (%s units waiting)" % len(self.waitq))

        # iterate on copy of waitq, as we manipulate the list during iteration.
        for unit in self.waitq[:] :

            uid = unit.uid
            ud  = unit.description

          # logger.debug ("examine unit  %s (%s cores)" % (uid, ud.cores))

            for pid in self.pilots :

              # logger.debug ("        pilot %s (%s caps, state %s)" \
              #            % (pid, self.pilots[pid]['state'], self.pilots[pid]['caps']))

                if  self.pilots[pid]['state'] in [ACTIVE] :

                    if  ud.cores <= self.pilots[pid]['caps'] :
                
                      # logger.debug ("        unit  %s fits on pilot %s" % (uid, pid))

                        # sanity check on unit state
                        if  unit.state not in [NEW] :
                            raise RuntimeError ("scheduler queue should only contain NEW units (%s)" % uid)

                        self.pilots[pid]['caps'] -= ud.cores
                        schedule['units'][unit]   = pid

                        # scheduled units are removed from the waitq
                        self.waitq.remove (unit)
                        self.runq.append  (unit)
                        break


                # unit was not scheduled...
                schedule['units'][unit] = None

            # print a warning if a unit cannot possibly be scheduled, ever
            can_handle_unit = False
            for pid in self.pilots :
                if  unit.description.cores <= self.pilots[pid]['cores'] :
                    can_handle_unit=True
                    break

            if  not can_handle_unit :
                logger.warning ('cannot handle unit %s with current set of pilots' % uid)

      # pprint.pprint (schedule)

        # tell the UM about the schedule
        self.manager.handle_schedule (schedule)

    # --------------------------------------------------------------------------




#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.scheduler
   :platform: Unix
   :synopsis: Scheduler implementations.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

from direct_submission import DirectSubmissionScheduler
from round_robin       import RoundRobinScheduler
from backfilling       import BackfillingScheduler

# -----------------------------------------------------------------------------
# Constants
SCHED_ROUND_ROBIN       = "round_robin"
SCHED_DIRECT_SUBMISSION = "direct_submission"
SCHED_BACKFILLING       = "backfilling"

# -----------------------------------------------------------------------------
# 
def get_scheduler(manager, name, session):
    """get_scheduler returns a scheduler object for 'name'.
    """
    if   name == SCHED_ROUND_ROBIN      : return RoundRobinScheduler       (manager, session)
    elif name == SCHED_DIRECT_SUBMISSION: return DirectSubmissionScheduler (manager, session)
    elif name == SCHED_BACKFILLING      : return BackfillingScheduler      (manager, session)
    else                                : raise RuntimeError("Scheduler '%s' doesn't exist." % name)




"""API layer tests
"""

import os
import sys
import radical.pilot
import unittest

import uuid
from copy import deepcopy
from radical.pilot.db import Session
from pymongo import MongoClient

# DBURL defines the MongoDB server URL and has the format mongodb://host:port.
# For the installation of a MongoDB server, refer to the MongoDB website:
# http://docs.mongodb.org/manual/installation/
DBURL = os.getenv("RADICAL_PILOT_DBURL")
if DBURL is None:
    print "ERROR: RADICAL_PILOT_DBURL (MongoDB server URL) is not defined."
    sys.exit(1)
    
DBNAME = os.getenv("RADICAL_PILOT_TEST_DBNAME")
if DBNAME is None:
    print "ERROR: RADICAL_PILOT_TEST_DBNAME (MongoDB database name) is not defined."
    sys.exit(1)


#-----------------------------------------------------------------------------
#
class Test_Session(unittest.TestCase):
    # silence deprecation warnings under py3

    def setUp(self):
        # clean up fragments from previous tests
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def tearDown(self):
        # clean up after ourselves 
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def failUnless(self, expr):
        # St00pid speling.
        return self.assertTrue(expr)

    def failIf(self, expr):
        # St00pid speling.
        return self.assertFalse(expr)

    #-------------------------------------------------------------------------
    #
    def test__session_create(self):
        """ Tests if creating a new session works as epxected.
        """
        for _ in range(1, 4):
            session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)
            
        client = MongoClient(DBURL)
        collections = client[DBNAME].collection_names()
        assert len(collections) == 4, "Wrong number of sessions in database"

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__session_reconnect(self):
        """ Tests if reconnecting to an existing session works as epxected.
        """
        session_ids = []
        for _ in range(1, 4):
            session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)
            session_ids.append(session.uid)

        for sid in session_ids:
            session_r = radical.pilot.Session(database_url=DBURL, database_name=DBNAME, session_uid=sid)
            assert session_r.uid == sid, "Session IDs don't match"

        session.close()




__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import radical.utils

import saga.attributes  as attributes
from radical.pilot.exceptions import * 

# -----------------------------------------------------------------------------
# Attribute description keys
JOB_MANAGER_ENDPOINT        = 'job_manager_endpoint'
FILESYSTEM_ENDPOINT         = 'filesystem_endpoint'
DEFAULT_QUEUE               = 'default_queue'
SPMD_VARIATION              = 'spmd_variation'
PYTHON_INTERPRETER          = 'python_interpreter'
PRE_BOOTSTRAP               = 'pre_bootstrap'
VALID_ROOTS                 = 'valid_roots'
BOOTSTRAPPER                = 'bootstrapper'
PILOT_AGENT                 = 'pilot_agent'
PILOT_AGENT_WORKER          = 'pilot_agent_worker'
GLOBAL_VIRTENV              = 'global_virtenv'
LRMS                        = 'lrms'
TASK_LAUNCH_METHOD          = 'task_launch_method'
MPI_LAUNCH_METHOD           = 'mpi_launch_method'
FORWARD_TUNNEL_ENDPOINT     = 'forward_tunnel_endpoint'
AGENT_MONGODB_ENDPOINT      = 'agent_mongodb_endpoint'
DEFAULT_REMOTE_WORKDIR      = 'default_remote_workdir'
DESCRIPTION                 = 'description'
NOTES                       = 'notes'
SCHEMAS                     = 'schemas'


VALID_KEYS = [JOB_MANAGER_ENDPOINT, FILESYSTEM_ENDPOINT, SCHEMAS,
              DEFAULT_QUEUE, SPMD_VARIATION, PYTHON_INTERPRETER, PRE_BOOTSTRAP, 
              VALID_ROOTS, BOOTSTRAPPER, PILOT_AGENT, PILOT_AGENT_WORKER,
              GLOBAL_VIRTENV, LRMS, TASK_LAUNCH_METHOD,
              MPI_LAUNCH_METHOD, FORWARD_TUNNEL_ENDPOINT, AGENT_MONGODB_ENDPOINT,
              DEFAULT_REMOTE_WORKDIR, NOTES, DESCRIPTION]

# -----------------------------------------------------------------------------
#
class ResourceConfig(attributes.Attributes):
    """A ResourceConfig object describes the internal configuration parameters 
    of a compute and / or storage resource. It can be passed to a 
    :class:`radical.pilot.PilotManager` to override or extend an existing 
    configuration set. 

    **Example**::

          rc = radical.pilot.ResourceConfig()
          rc.name                 = "ec2.my_cluster"
          rc.job_manager_endpoint = "ssh://23.23.23.23/"
          rc.filesystem_endpoint  = "sftp://23.23.23.23"
          rc.default_queue        = "batch"
          rc.python_interpreter   = "/opt/python/2.7.6/bin/python"
          rc.pre_bootstrap        = "module load mpi"
          rc.valid_roots          = ["/home", "/work"]
          rc.bootstrapper         = "default_bootstrapper.sh"

          pmgr = radical.pilot.PilotManager(session=session)
          pmgr.add_resource_config(rc)

          // [...]

          pd = radical.pilot.ComputePilotDescription()
          pd.resource = "epsrc.archer"

          // [...]

          pmgr.submit_pilots(pd)

    .. data:: name

       [Type: `string`] [**`mandatory`**] A unique name for this configuration. 

    .. data:: remote_job_manager_endpoint

       [Type: `string`] [optional] TODO

    .. data:: remote_filesystem_endpoint

       [Type: `string`] [optional] TODO

    .. data:: local_job_manager_endpoint

       [Type: `string`] [optional] TODO

    .. data:: local_filesystem_endpoint

       [Type: `string`] [optional] TODO

    .. data:: default_queue

       [Type: `string`] [optional] TODO

    .. data:: spmd_variation

       [Type: `string`] [optional] TODO

    .. data:: python_interpreter

       [Type: `string`] [optional] TODO

    .. data:: pre_bootstrap

       [Type: `string`] [optional] TODO

    .. data:: valid_roots

       [Type: `string`] [optional] TODO

    .. data:: bootstrapper

       [Type: `string`] [optional] TODO

    .. data:: pilot_agent

       [Type: `string`] [optional] TODO

    .. data:: global_virtenv

       [Type: `string`] [optional] TODO

    .. data:: lrms

       [Type: `string`] [optional] TODO

    """

    # -------------------------------------------------------------------------
    #
    @staticmethod 
    def from_file(filename, entry_name=None):
      """Reads a resource configuration JSON file from the URL provided and 
         returns a list of one or more ResourceConfig objects.
      """
      rcfgs = dict()

      try:
          rcf_dict = radical.utils.read_json_str (filename)
          rcf_name = str(os.path.basename (filename))

          if  rcf_name.endswith ('.json'):
              rcf_name = rcf_name[0:-5]

          if  'aliases' in rcf_dict:
              return None

          for res_name, cfg in rcf_dict.iteritems():

              # create config from resource section
              cls = ResourceConfig(cfg)

              # make sure all keys are initialized
              for key in VALID_KEYS:
                  if not key in cls :
                      cls[key] = None
            
              rcfgs["%s.%s" % (rcf_name, res_name)] = cls

      except ValueError, err:
          raise BadParameter("Couldn't parse resource configuration file '%s': %s." % (filename, str(err)))

      return rcfgs


    # -------------------------------------------------------------------------
    #
    def __init__(self, seeding_dict=None):
        """Optionally take a seeding dict to populate the values.
        """

        # initialize attributes
        attributes.Attributes.__init__(self, seeding_dict)

        # set attribute interface properties
        self._attributes_extensible  (True)
        self._attributes_camelcasing (True)

        self._attributes_register(JOB_MANAGER_ENDPOINT,    None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(FILESYSTEM_ENDPOINT,     None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(DEFAULT_QUEUE,           None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(SPMD_VARIATION,          None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(PYTHON_INTERPRETER,      None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(PRE_BOOTSTRAP,           None, attributes.STRING, attributes.VECTOR, attributes.WRITEABLE)
        self._attributes_register(VALID_ROOTS,             None, attributes.STRING, attributes.VECTOR, attributes.WRITEABLE)
        self._attributes_register(BOOTSTRAPPER,            None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(PILOT_AGENT,             None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(PILOT_AGENT_WORKER,      None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(GLOBAL_VIRTENV,          None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(LRMS,                    None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(TASK_LAUNCH_METHOD,      None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(MPI_LAUNCH_METHOD,       None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(FORWARD_TUNNEL_ENDPOINT, None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(AGENT_MONGODB_ENDPOINT,  None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(DEFAULT_REMOTE_WORKDIR,  None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(DESCRIPTION,             None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)
        self._attributes_register(NOTES,                   None, attributes.STRING, attributes.SCALAR, attributes.WRITEABLE)


    # -------------------------------------------------------------------------
    #
    def __str__(self):
        """Returns a string representation of the object.
        """
        return str(self.as_dict())



#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.scheduler.DirectSubmissionScheduler
   :platform: Unix
   :synopsis: A single-pilot, direct submission scheduler.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os

from radical.pilot.utils.logger        import logger
from radical.pilot.scheduler.interface import Scheduler 

# -----------------------------------------------------------------------------
# 
class DirectSubmissionScheduler(Scheduler):
    """DirectSubmissionScheduler implements a single-pilot 'pass-through' 
    scheduling algorithm.
    """

    # -------------------------------------------------------------------------
    #
    def __init__(self, manager, session):
        """Le constructeur.
        """

        self.manager = manager
        self.session = session
        self.pilots  = dict()

        logger.info("Loaded scheduler: %s." % self.name)

    # -------------------------------------------------------------------------
    #
    def pilot_added (self, pilot) :

        pid = pilot.uid

        self.pilots[pid] = dict()
        self.pilots[pid]['resource'] = pilot.resource
        self.pilots[pid]['sandbox']  = pilot.sandbox


    # -------------------------------------------------------------------------
    #
    def pilot_removed (self, pid) :

        if  not pid in self.pilots :
            raise RuntimeError ('cannot remove unknown pilot (%s)' % pid)

        del self.pilots[pid]


    # -------------------------------------------------------------------------
    #
    def schedule(self, units):

        pilot_ids = self.pilots.keys ()

        if not len (pilot_ids):
            raise RuntimeError ('Unit scheduler cannot operate on empty pilot set')

        if len (pilot_ids) > 1:
            raise RuntimeError ('Direct Submission only works for a single pilot!')
        
        schedule           = dict()
        schedule['units']  = dict()
        schedule['pilots'] = self.pilots

        for unit in units:

            schedule['units'][unit] = pilot_ids[0]

        return schedule

